import enum
import traceback
from typing import Any, List, Tuple

from google.protobuf.descriptor import Descriptor, FieldDescriptor


class ProtoType(enum.Enum):
    """Protobuf enum types

    Notes
    -----
        For protobuf type enums see:
        https://developers.google.com/protocol-buffers/docs/reference/cpp/google.protobuf.descriptor.pb
    """

    DOUBLE = 1
    FLOAT = 2
    INT64 = 3
    UINT64 = 4
    INT32 = 5
    FIXED64 = 6
    FIXED32 = 7
    BOOL = 8
    STRING = 9
    GROUP = 10
    MESSAGE = 11
    BYTES = 12
    UINT32 = 13
    ENUM = 14
    SFIXED32 = 15
    SFIXED64 = 16
    SINT32 = 17
    SINT64 = 18


class SqliteType(enum.Enum):
    """Enum for sqlite3 datatypes"""

    REAL = "READ"
    INTEGER = "INTEGER"
    STRING = "STRING"
    BLOB = "BLOB"


def get_create_table_cmds_for_enum(enum_pb2_class: Any) -> List[str]:
    """Get the commands to create a table for a protobuf enum

    Parameters
    ----------
    enum_pb2_class : Any
        protobuf msg class

    Returns
    -------
    cmds : List[str]
        sqlite3 commands for creating the table
        and inserting the enum values
    """
    table_name = enum_pb2_class.DESCRIPTOR.name

    cmds = []
    cmds.append(
        get_create_table_cmd(
            table_name,
            [
                ("name", SqliteType.STRING),
                ("value", SqliteType.INTEGER),
            ],
            ["name"],
        )
    )

    for name, value in enum_pb2_class.items():
        cmds.append(
            f"INSERT OR REPLACE INTO {table_name} "
            + "(name, value) "
            + f"VALUES('{name}',{value})"
        )

    return cmds


def field_descriptor_is_list(field_descriptor: FieldDescriptor) -> bool:
    """Checks if a proto field is a list or not

    Parameters
    ----------
    field_descriptor : FieldDescriptor
        descriptor of a field within a proto msg

    Returns
    -------
    is_list : bool
    """
    return isinstance(field_descriptor.default_value, list)


def sql_type_from_proto_type(
    proto_type: int,
    message_name: str,
    field_name: str,
) -> SqliteType:
    """Translate protobuf types into sqlite types

    Parameters
    ----------
    proto_type : int
        protobuf type enum
    message_name : str
        message name for better error messages
    field_name : str
        field name for better error messages

    Returns
    -------
    sqlite_type : SqliteType
        enum type of sqlite type

    Notes
    -----
        For protobuf type enums see:
        https://developers.google.com/protocol-buffers/docs/reference/cpp/google.protobuf.descriptor.pb
    """
    if proto_type in (ProtoType.DOUBLE.value, ProtoType.FLOAT.value):
        return SqliteType.REAL

    if proto_type in (
        ProtoType.INT64.value,
        ProtoType.UINT64.value,
        ProtoType.INT32.value,
        ProtoType.FIXED64.value,
        ProtoType.FIXED32.value,
        ProtoType.BOOL.value,
        ProtoType.UINT32.value,
        ProtoType.SFIXED32.value,
        ProtoType.SFIXED64.value,
        ProtoType.SINT32.value,
        ProtoType.SINT64.value,
    ):
        return SqliteType.INTEGER

    if proto_type == ProtoType.STRING.value:
        return SqliteType.STRING

    if proto_type == ProtoType.BYTES.value:
        return SqliteType.BLOB

    err_msg = (
        f"{message_name}.{field_name} "
        + f"has an unknown protobuf type enum: {proto_type}"
    )
    raise RuntimeError(err_msg)


def get_proto_fields_with_types(
    pb_descriptor: Descriptor,
) -> List[Tuple[str, SqliteType]]:
    """Get the proto field names of a descriptor with types

    Parameters
    ----------
    pb_descriptor: descriptor.Descriptor
        descriptor from protobuf basically describing a message

    Returns
    -------
    fields : List[Tuple[str, SqliteType]]
        fields with (name, type) tuple
    """
    try:
        return [
            (
                field.name,
                sql_type_from_proto_type(
                    field.type, pb_descriptor.full_name, field.name
                ),
            )
            for field in pb_descriptor.fields
        ]
    except RuntimeError:
        err_msg = traceback.format_exc() + f"\nError in {pb_descriptor.full_name}"
        raise RuntimeError(err_msg) from None


def get_sqlite_fields_for_insertion_from_pb2(pb_descriptor: Descriptor) -> str:
    """Converts proto fields from a proto descriptor into sqlite fields
    for table insertion

    Parameters
    ----------
    pb_descriptor: descriptor.Descriptor
        descriptor from protobuf basically describing a message

    Returns
    -------
    fields : str
        string usable for table insertion
    """
    field_name_and_type = get_proto_fields_with_types(pb_descriptor)
    field_names_prepped = (f"'{name}'" for (name, _) in field_name_and_type)
    return ",".join(field_names_prepped)


def get_create_table_cmd(
    table_name: str,
    attributes: List[Tuple[str, SqliteType]],
    primary_key_attribute_names: List[str],
) -> str:
    """Get the command to create a sqlite table from the
    definition of the attributes

    Parameters
    ----------
    table_name : str
        name of the table
    attributes : Dict[str, SqliteType]
        attributes to store in the sqlite table
    key_attribute_names : List[str]
        names of the attributes used as keys

    Returns
    -------
    create_tbl_cmd : str
        command for sqlite to create a table
    """

    format_str = "{0} {1}"
    name_and_type_as_str = (
        format_str.format(attribute_name, attribute_type.value)
        for attribute_name, attribute_type in attributes
    )
    attribute_string = ", ".join(name_and_type_as_str)

    return (
        f"CREATE TABLE IF NOT EXISTS {table_name}("
        + attribute_string
        + ",PRIMARY KEY("
        + ",".join(primary_key_attribute_names)
        + ")"
        + ")"
    )


def sqlite_create_tbl_cmd_from_pb2(
    table_name: str,
    meta_attributes: List[Tuple[str, SqliteType]],
    primary_key_attribute_names: List[str],
    pb_descriptor: Descriptor,
) -> str:
    """Get the table creation command from a proto msg

    Parameters
    ----------
    table_name : str
        name of the table
    meta_attributes : List[Tuple[str, SqliteType]]
        additional attributes to store in the sqlite table
    pb_descriptor : descriptor.Descriptor
        proto message to be stored

    Returns
    -------
    create_tbl_cmd : str
        command for sqlite to create a table
    """

    pb2_attributes = get_proto_fields_with_types(pb_descriptor)

    # enforce unique naming to be safe
    set_meta_names = set(name for name, _ in meta_attributes)
    set_pb2_names = set(name for name, _ in pb2_attributes)
    duplicate_names = set_meta_names.intersection(set_pb2_names)
    if duplicate_names:
        err_msg = (
            "The following meta names intersect " + "with protobuf message names: {0}"
        )
        raise ValueError(err_msg.format(", ".join(duplicate_names)))

    set_primary_names = set(primary_key_attribute_names)
    set_all_attribute_names = set_meta_names.union(set_pb2_names)
    missing_attributes = set_primary_names - set_all_attribute_names
    if missing_attributes:
        err_msg = (
            "The following primary key attribute names do"
            + " neither exist in the proto file nor the additionally"
            + " specified attributes: "
            + ", ".join(missing_attributes)
        )

    return get_create_table_cmd(
        table_name=table_name,
        attributes=meta_attributes + pb2_attributes,
        primary_key_attribute_names=primary_key_attribute_names,
    )



from ..protobuf.generated.config_pb2 import ChiaTeaConfig


def get_credentials_key(
    is_testing: bool,
    config: ChiaTeaConfig,
) -> bytes:
    """Get the private key from the config

    Parameters
    ---------
    is_testing : bool
        whether testing is being performed
    config : ChiaTeaConfig
        config to read filepath for cert and key

    Returns
    -------
    key : str
        key file as string
    """

    if is_testing:
        return b""

    key_filepath = config.monitoring.auth.key_filepath

    with open(key_filepath, "r", encoding="utf8") as fp:
        key = fp.read().encode("utf8")

    return key


def get_credentials_cert(is_testing: bool, config: ChiaTeaConfig) -> bytes:
    """Get the certificates from the config

    Parameters
    ---------
    is_testing : bool
        whether testing is being performed
    config : ChiaTeaConfig
        config to read filepath for cert and key

    Returns
    -------
    cert : str
        cert file as string
    """

    if is_testing:
        return b""

    cert_filepath = config.monitoring.auth.cert_filepath

    with open(cert_filepath, "r", encoding="utf8") as fp:
        cert = fp.read().encode("utf8")

    return cert



import importlib
import sqlite3
from collections.abc import Iterable
from typing import Any, Callable, Dict, List, Tuple

from google.protobuf.descriptor import Descriptor
from google.protobuf.message import Message

from ..generated.computer_info_pb2 import ADD, DELETE, NONE, UPDATE, UpdateEvent
from .generic import (
    ProtoType,
    SqliteType,
    field_descriptor_is_list,
    get_proto_fields_with_types,
    sqlite_create_tbl_cmd_from_pb2,
)

TableInsertionFunc = Callable[[sqlite3.Cursor, Any, Dict[str, Any]], None]
TableDeletionFunc = Callable[[sqlite3.Cursor, Any, int], None]


EVENT_TABLE_META_ATTRIBUTES = [
    ("timestamp", SqliteType.REAL),
    ("machine_id", SqliteType.INTEGER),
    ("event_type", SqliteType.INTEGER),
]

STATE_TABLE_META_ATTRIBUTES = [
    ("machine_id", SqliteType.INTEGER),
]


def sqlite_create_event_tbl_cmd_from_pb2(
    pb_descriptor: Descriptor,
    meta_attributes: List[Tuple[str, SqliteType]],
    primary_key_names: List[str],
    optional_primary_key_names: List[str],
) -> str:
    """Get a sqlite table creation cmd (if not exists)
    from a proto message descriptor

    Parameters
    ----------
    pb_descriptor: Descriptor
        descriptor from protobuf basically describing a message
    meta_attributes : List[Tuple[str, SqliteType]]
        meta attributes to store besides proto data
    primary_key_names : List[str]
        var names used for the primary key
    optional_primary_key_names : List[str]
        optional primary keys which are added if they are
        contained within the proto msg or meta_attributes
    Returns
    -------
    cmd : str
        command for execution
    """
    table_name = pb_descriptor.name + "Events"

    all_var_names = [field.name for field in pb_descriptor.fields] + [
        name for name, _ in meta_attributes
    ]

    for key_name in primary_key_names:
        if key_name not in all_var_names:
            err_msg = "primary key attribute '{0}' does not match any var: {1}"
            raise ValueError(err_msg.format(key_name, all_var_names))

    for additional_name in optional_primary_key_names:
        for field in pb_descriptor.fields:
            if field.name == additional_name:
                primary_key_names = [
                    *primary_key_names,
                    additional_name,
                ]

    return sqlite_create_tbl_cmd_from_pb2(
        table_name=table_name,
        meta_attributes=meta_attributes,
        primary_key_attribute_names=primary_key_names,
        pb_descriptor=pb_descriptor,
    )


def sqlite_create_state_tbl_cmd_from_pb2(
    pb_descriptor: Descriptor,
    meta_attributes: List[Tuple[str, SqliteType]],
    primary_key_names: List[str],
    optional_primary_key_names: List[str],
) -> str:
    """Get a sqlite table creation cmd (if not exists)
    from a proto message descriptor

    Parameters
    ----------
    pb_descriptor: Descriptor
        descriptor from protobuf basically describing a message
    meta_attributes : List[Tuple[str, SqliteType]]
        meta attributes to store besides proto data
    primary_key_names : List[str]
        var names used for the primary key
    optional_primary_key_names : List[str]
        optional primary keys which are added if they are
        contained within the proto msg or meta_attributes

    Returns
    -------
    cmd : str
        command for execution
    """
    table_name = pb_descriptor.name

    all_var_names = [field.name for field in pb_descriptor.fields] + [
        name for name, _ in meta_attributes
    ]

    for key_name in primary_key_names:
        if key_name not in all_var_names:
            err_msg = "primary key attribute '{0}' does not match any var: {1}"
            raise ValueError(err_msg.format(key_name, all_var_names))

    for additional_name in optional_primary_key_names:
        for field in pb_descriptor.fields:
            if field.name == additional_name:
                primary_key_names = [
                    *primary_key_names,
                    additional_name,
                ]

    return sqlite_create_tbl_cmd_from_pb2(
        table_name=table_name,
        meta_attributes=meta_attributes,
        primary_key_attribute_names=primary_key_names,
        pb_descriptor=pb_descriptor,
    )


def sqlite_insert_into_table_fun_from_pb2(
    table_name: str,
    meta_attribute_names: List[str],
    pb_descriptor: Descriptor,
    fields_to_ignore: List[str],
) -> TableInsertionFunc:
    """Create a function to insert a proto msg into sqlite
    from a pb2 message descriptor

    Parameters
    ----------
    table_name : str
        name of the sqlite table
    meta_attribute_names : List[str]
        names of the meta attributes
    pb_descriptor: Descriptor
        descriptor from protobuf basically describing a message
    fields_to_ignore : List[str]
        fields to ignore

    Returns
    -------
    function : Callable[[str, int, Any, sqlite.Cursor], None]
        function to insert a proto message into sqlite3
    """
    field_name_and_type = get_proto_fields_with_types(pb_descriptor)
    field_names = [
        field_name
        for field_name, _ in field_name_and_type
        if field_name not in fields_to_ignore
    ]

    all_attributes = meta_attribute_names + field_names
    all_names = ",".join(all_attributes)

    insertion_cmd = (
        f"INSERT OR REPLACE INTO {table_name} ("
        + all_names
        + ") "
        + "VALUES("
        + ",".join("?" * (len(all_attributes)))
        + ")"
    )

    def _table_insertion_function(
        sql_cursor: sqlite3.Cursor,
        pb_message: Message,
        meta_attributes: Dict[str, Any],
    ):
        if pb_message is None:
            return

        pb_message_iter = (
            (pb_message,) if not isinstance(pb_message, Iterable) else pb_message
        )

        # this must not fail
        meta_attribute_values = (
            meta_attributes[attribute_name] for attribute_name in meta_attribute_names
        )

        attributes = []
        for pb_entry in pb_message_iter:
            attributes.append(
                (
                    *meta_attribute_values,
                    *get_pb2_attributes_as_list(
                        pb_entry,
                        pb_descriptor,
                        fields_to_ignore=fields_to_ignore,
                    ),
                )
            )

        sql_cursor.executemany(
            insertion_cmd,
            attributes,
        )

    return _table_insertion_function


def sqlite_delete_in_table_fun_from_pb2(
    table_name: str,
    meta_attribute_names: List[str],
    pb_descriptor: Descriptor,
) -> TableDeletionFunc:
    """Create a function to delete a proto msg in a sqlite table
    from a pb2 message descriptor

    Parameters
    ----------
    table_name : str
        name of the sqlite table
    meta_attribute_names: List[str]
        names of the meta attributes used to
        identify which entry to delete
    pb_descriptor: Descriptor
        descriptor from protobuf describing a message

    Returns
    -------
    function : TableDeletionFunc
        function to insert a proto message into sqlite3
    """

    deletion_attributes = list(meta_attribute_names)

    # pylint: disable=unused-argument
    def _get_pb2_id(pb2_message) -> Tuple[Any, ...]:
        return tuple()

    has_id = any(field.name == "id" for field in pb_descriptor.fields)
    if has_id:
        deletion_attributes.append("id")

        def _get_pb2_id(pb_message: Message) -> Tuple[Any, ...]:  # noqa: F811
            return (pb_message.id,)

    deletion_cmd = (
        f"DELETE FROM {table_name} "
        + " WHERE "
        + " AND ".join(f"{name}=?" for name in deletion_attributes)
    )

    def _table_insertion_function(
        sql_cursor: sqlite3.Cursor,
        pb_message: Message,
        meta_attributes: Dict[str, Any],
    ):
        if pb_message is None:
            return

        pb_message_iter = (
            (pb_message,) if not isinstance(pb_message, Iterable) else pb_message
        )

        # this must not fail
        meta_attribute_values: Tuple[Any, ...] = tuple(
            meta_attributes[attribute_name] for attribute_name in meta_attribute_names
        )

        attributes = []
        for pb_entry in pb_message_iter:

            attributes.append(
                (
                    *meta_attribute_values,
                    *_get_pb2_id(
                        pb_entry,
                    ),
                )
            )

        sql_cursor.executemany(
            deletion_cmd,
            attributes,
        )

    return _table_insertion_function


def get_pb2_attributes_as_list(
    pb_msg: Message,
    pb_descriptor: Descriptor,
    fields_to_ignore: List[str],
) -> List[Any]:
    """Extract the attributes of a proto message into a list

    Parameters
    ----------
    pb_msg : Any
        protobuf message (from pb_descriptor)
    pb_descriptor: Descriptor
        descriptor from protobuf basically describing a message
    fields_to_ignore : List[str]
        which fields to ignore

    Returns
    -------
    attributes : List[Any]
        message attributes as list
    """
    return [
        getattr(pb_msg, field.name)
        for field in pb_descriptor.fields
        if field.name not in fields_to_ignore
    ]


def get_table_insertion_function_for_nested_messages(
    table_suffix: str,
    meta_attribute_names: List[str],
    pb_descriptor: Descriptor,
) -> TableInsertionFunc:
    """Get command to insert nested submessages into their tables

    Parameters
    ----------
    table_suffix : str
        custom suffix for the table
    meta_attribute_names : List[str]
        names of the meta attributes
    pb_descriptor: Descriptor
        descriptor from protobuf basically a message

    Returns
    -------
    insertion_funtion : TableInsertionFunc
        function to insert the main message into the database

    Notes
    -----
        Requires to run `get_event_table_creation_cmds_for_nested_messages`
        first.
    """
    msgs_insertion_functions = {
        field.number: sqlite_insert_into_table_fun_from_pb2(
            field.message_type.name + table_suffix,
            meta_attribute_names,
            field.message_type,
            [],
        )
        for field in pb_descriptor.fields
        if field.type == ProtoType.MESSAGE.value
    }

    def _insertion_function(
        sql_cursor: sqlite3.Cursor,
        pb_message: Any,
        meta_attributes: Dict[str, Any],
    ):

        field_name, pb_submessage = get_update_even_data(pb_message)
        field_number = pb_descriptor.fields_by_name[field_name].number

        fun = msgs_insertion_functions[field_number]
        fun(
            sql_cursor,
            pb_submessage,
            meta_attributes,
        )

    return _insertion_function


def get_event_table_insertion_cmds_for_nested_messages(
    pb_descriptor: Descriptor,
) -> TableInsertionFunc:
    """Get command to insert nested submessages into their tables

    Parameters
    ----------
    pb_descriptor: Descriptor
        descriptor from protobuf basically a message

    Returns
    -------
    insertion_funtion : TableInsertionFunc
        function to insert the main message into the database

    Notes
    -----
        Requires to run `get_event_table_creation_cmds_for_nested_messages`
        first.
    """
    return get_table_insertion_function_for_nested_messages(
        table_suffix="Events",
        meta_attribute_names=[name for name, _ in EVENT_TABLE_META_ATTRIBUTES],
        pb_descriptor=pb_descriptor,
    )


StateModificationFun = Callable[
    [
        sqlite3.dbapi2.Cursor,
        Any,
        Dict[str, Any],
        int,
    ],
    Any,
]


def get_state_table_modification_fun_for_nested_messages(
    pb_descriptor: Descriptor,
) -> StateModificationFun:
    """Get command to insert nested submessages into their tables

    Parameters
    ----------
    pb_descriptor: Descriptor
        descriptor from protobuf basically a message

    Returns
    -------
    insertion_funtion : TableInsertionFunc
        function to insert the main message into the database
        Requires to run `get_state_table_creation_cmds_for_nested_messages`
        first.
    """
    table_suffix = ""
    meta_attribute_names = [name for name, _ in STATE_TABLE_META_ATTRIBUTES]

    msgs_insertion_functions = {
        field.number: sqlite_insert_into_table_fun_from_pb2(
            field.message_type.name + table_suffix,
            meta_attribute_names,
            field.message_type,
            [],
        )
        for field in pb_descriptor.fields
        if field.type == ProtoType.MESSAGE.value
    }

    msgs_deletion_functions = {
        field.number: sqlite_delete_in_table_fun_from_pb2(
            field.message_type.name + table_suffix,
            meta_attribute_names,
            field.message_type,
        )
        for field in pb_descriptor.fields
        if field.type == ProtoType.MESSAGE.value
    }

    def _deletion_function(
        sql_cursor: sqlite3.Cursor,
        pb_message: Any,
        meta_attributes: Dict[str, Any],
        event_type: int,
    ):

        if event_type == NONE:
            return

        field_name, pb_submessage = get_update_even_data(pb_message)
        field_number = pb_descriptor.fields_by_name[field_name].number

        if event_type in (ADD, UPDATE):
            fun = msgs_insertion_functions[field_number]
        elif event_type == DELETE:
            fun = msgs_deletion_functions[field_number]

        fun(
            sql_cursor,
            pb_submessage,
            meta_attributes,
        )

    return _deletion_function


def get_function_to_retrieve_pb2_from_sqlite_db(
    table_suffix: str,
    pb_class: Any,
    key_names_and_ops: List[Tuple[str, str]],
) -> Callable[[sqlite3.Cursor, List[Any]], Tuple[List[Message], List[Dict[str, Any]]]]:
    """Get a function to read entries from the sqlite db
    and convert them abck into pb2 messages

    Parameters
    ----------
    table_suffix : str
        suffix for the table name (prefix is proto name)
    pb_class : Any
        protobuf class used to create a message
    key_names_and_ops : List[Tuple[str, str]]
        name of the columns used for retrieval and the operation
        associated with it such as '=' or '<'.

    Returns
    -------
    func : Callable[[sqlite3.Cursor, List[Any]], List[Message]]
        function for retrieving proto messages from the sqlite db
    """

    pb_descriptor = pb_class.DESCRIPTOR
    table_name = pb_descriptor.name + table_suffix

    selectors = " AND ".join(
        f"{name}{operation}?" for name, operation in key_names_and_ops
    )

    comparison_word = " WHERE " if selectors else ""

    cmd = f"SELECT * FROM {table_name}" + comparison_word + selectors

    def _get_state_as_pb(
        cursor: sqlite3.Cursor,
        *key_values: List[Any],
    ) -> Tuple[List[Message], List[Dict[str, Any]]]:

        cursor.execute(cmd, key_values)
        rows = cursor.fetchall()

        # this contains the names but if nothing is
        # found this is none
        description = cursor.description or tuple()

        entry_names = tuple(entry[0] for entry in description)

        pb_messages: List[Message] = []
        attributes = []
        for row in rows:
            pb_msg = pb_class()
            msg_attributes = {}
            for name, value in zip(entry_names, row):
                if hasattr(pb_msg, name):
                    setattr(pb_msg, name, value)
                else:
                    msg_attributes[name] = value
            pb_messages.append(pb_msg)
            attributes.append(msg_attributes)

        return pb_messages, attributes

    return _get_state_as_pb


def get_fun_to_collect_pb2_messages_for_nested_submessages(
    pb_class: Any,
    key_names_and_ops: List[Tuple[str, str]],
    table_suffix: str,
) -> Callable[[sqlite3.Cursor, List[Any]], Message]:
    """Get a function to get a proto message with nested messages

    Parameters
    ----------
    pb_class : Any
        protobuf class used to create messages
    key_names_and_ops : List[Tuple[str, str]]
        name of the columns used for retrieval and the operation
        associated with it such as '=' or '<'.
    table_suffix : str
        suffix for the table name. Prefix is the proto class name

    Returns
    -------
    fun : Callable[[sqlite3.Cursor, List[Any]], Message]
        function to retrieve the proto message from the speified
        primary key attributes
    """

    pb_descriptor = pb_class.DESCRIPTOR

    msgs_retrieve_functions = {}
    msg_is_iterable = {}
    for field in pb_descriptor.fields:

        if field.type != ProtoType.MESSAGE.value:
            continue

        # We can only import a proto class here
        # when the module path is set correctly in
        # the proto file.

        # get the class from the proto msg descriptor
        message_descriptor = field.message_type
        module_name, class_name = message_descriptor.full_name.rsplit(".", 1)
        module = importlib.import_module(module_name)
        class_constructor = getattr(module, class_name)

        msgs_retrieve_functions[
            field.number
        ] = get_function_to_retrieve_pb2_from_sqlite_db(
            table_suffix=table_suffix,
            pb_class=class_constructor,
            key_names_and_ops=key_names_and_ops,
        )
        msg_is_iterable[field.number] = field_descriptor_is_list(field)

    def _retrieve_function(
        cursor: sqlite3.Cursor,
        primary_keys: List[Any],
    ):

        kwargs = {}
        for field in pb_descriptor.fields:

            if field.type != ProtoType.MESSAGE.value:
                continue

            fun = msgs_retrieve_functions[field.number]
            messages, attributes = fun(cursor, primary_keys)

            if msg_is_iterable[field.number]:
                kwargs[field.name] = messages
            else:
                kwargs[field.name] = messages[0] if messages else None

            if attributes:
                for attr_name, attr_value in attributes[0].items():
                    if attr_name in pb_descriptor.fields_by_name:
                        kwargs[attr_name] = attr_value

        return pb_class(**kwargs)

    return _retrieve_function


def get_fun_to_collect_latest_update_events_from_db(
    pb_class: Any,
    key_names_and_ops: List[Tuple[str, str]],
    table_suffix: str,
) -> Callable[[sqlite3.Cursor, Dict[str, Any]], Dict[int, List[UpdateEvent]]]:
    """Get a function to get a proto message with nested messages

    Parameters
    ----------
    pb_class : Any
        protobuf class used to create messages
    key_names_and_ops : List[Tuple[str, str]]
        name of the columns used for retrieval and the operation
        associated with it such as '=' or '<' as string.
    table_suffix : str
        suffix for the table name. Prefix is the proto class name

    Returns
    -------
    fun : Callable[[sqlite3.Cursor, Dict[str, Any]], Dict[int, List[UpdateEvent]]]
        function to retrieve the proto message from the speified
        primary key attributes
    """

    pb_descriptor = pb_class.DESCRIPTOR

    msgs_retrieve_functions = {}
    for field in pb_descriptor.fields:

        if field.type != ProtoType.MESSAGE.value:
            continue

        # We can only import a proto class here
        # when the module path is set correctly in
        # the proto file.

        # get the class from the proto msg descriptor
        message_descriptor = field.message_type
        module_name, class_name = message_descriptor.full_name.rsplit(".", 1)
        module = importlib.import_module(module_name)
        class_constructor = getattr(module, class_name)

        msgs_retrieve_functions[
            field.number
        ] = get_function_to_retrieve_pb2_from_sqlite_db(
            table_suffix=table_suffix,
            pb_class=class_constructor,
            key_names_and_ops=key_names_and_ops,
        )

    def _retrieve_function(
        cursor: sqlite3.Cursor,
        *key_values: List[Any],
    ) -> Dict[int, List[UpdateEvent]]:

        collected_events: Dict[int, List[UpdateEvent]] = {}

        for field in pb_descriptor.fields:

            if field.type != ProtoType.MESSAGE.value:
                continue

            fun = msgs_retrieve_functions[field.number]
            messages, attributes = fun(cursor, *key_values)

            for pb_submsg, attribute_dict in zip(messages, attributes):
                instance = pb_class(**{field.name: pb_submsg})

                for name, value in attribute_dict.items():
                    if hasattr(instance, name):
                        setattr(instance, name, value)

                machine_id = attribute_dict["machine_id"]

                event_list = collected_events.get(machine_id, [])
                event_list.append(instance)
                collected_events[machine_id] = event_list

        return collected_events

    return _retrieve_function


def get_fun_to_update_db_from_pb2_class(
    pb_class: Any,
    key_names_and_ops: List[Tuple[str, str]],
    table_suffix: str,
    attribute_names: List[str],
) -> Callable[[sqlite3.Cursor, Message, Dict[str, Any], Dict[str, Any]], None]:
    """Get a function to update database entries

    Parameters
    ----------
    pb_class : Any
        protobuf class used to create messages
    key_names_and_ops : List[Tuple[str, str]]
        name of the columns used for retrieval and the operation
        associated with it such as '=' or '<' as string.
    table_suffix : str
        suffix for the table name. Prefix is the proto class name
    attribute_names : List[str],
        additional attributes to modify

    Returns
    -------
    _update_table_fun : Callable[[sqlite3.Cursor, Message, Dict[str, Any], Dict[str, Any]], None]
        function to update table entries
    """

    pb_descriptor = pb_class.DESCRIPTOR
    table_name = pb_descriptor.name + table_suffix

    setter_fields = [f"{field.name}=?" for field in pb_descriptor.fields]
    all_fields = attribute_names + setter_fields

    selectors = " AND ".join(
        f"{name}{operation}?" for name, operation in key_names_and_ops
    )
    comparison_word = " WHERE " if selectors else ""

    cmd = (
        f"UPDATE {table_name} "
        + "SET "
        + ",".join(all_fields)
        + comparison_word
        + selectors
    )

    def _update_table_fun(
        cursor: sqlite3.Cursor,
        pb_message: Message,
        attributes: Dict[str, Any],
        ops_values: List[Any],
    ):

        if pb_message is None:
            return

        pb_message_iter = (
            (pb_message,) if not isinstance(pb_message, Iterable) else pb_message
        )

        # this must not fail
        attribute_values_as_tuple: Tuple[Any, ...] = tuple(
            attributes[attribute_name] for attribute_name in attribute_names
        )

        values_to_insert_into_cmd = []
        for pb_entry in pb_message_iter:
            values_to_insert_into_cmd.append(
                (
                    *attribute_values_as_tuple,
                    *get_pb2_attributes_as_list(
                        pb_entry,
                        pb_descriptor,
                        fields_to_ignore=[],
                    ),
                    *ops_values,
                )
            )

        cursor.executemany(cmd, values_to_insert_into_cmd)

    return _update_table_fun


def get_update_even_data(update_event: UpdateEvent) -> Tuple[str, Message]:
    """Get the data of an update event

    Parameters
    ----------
    update_event : UpdateEvent
        event containing data update

    Returns
    -------
    name : str
        submessage variable name
    pb_msg : Message
        update event data
    """
    field_name = update_event.WhichOneof("event_data")
    sub_msg = getattr(
        update_event,
        field_name,
    )
    return field_name, sub_msg



import asyncio
import unittest
from datetime import datetime

from ..utils.testing import async_test
from .ChiaWatchdog import ChiaWatchdog
from .logfile.FarmerHarvesterLogfile import FarmerHarvesterLogfile
from .logfile.line_checks import ActionFarmedUnfinishedBlock


class TestChiaWatchdog(unittest.TestCase):
    @async_test
    async def test_readiness(self):
        dog = ChiaWatchdog("", "")

        async def assert_not_ready():
            with self.assertRaises(asyncio.TimeoutError):
                await asyncio.wait_for(dog.ready(), timeout=0.5)

        dog.set_chia_logfile_is_ready()
        await assert_not_ready()
        dog.set_madmax_logfile_is_ready()
        await assert_not_ready()
        dog.farmer_service.is_ready = True
        await assert_not_ready()
        dog.harvester_service.is_ready = True
        await assert_not_ready()
        dog.wallet_service.is_ready = True
        # raising an exception is an error case here
        await asyncio.wait_for(dog.ready(), timeout=0.5)

    def test_reward_found(self):
        action1 = ActionFarmedUnfinishedBlock()
        node_id = "65322a31ad01f3aa3fc04f3e43231d35c3a1ddd4"
        ip_address = "57.22.39.97"
        timestamp_str = "2021-05-26T09:37:13.872"
        timestamp = datetime.fromisoformat(timestamp_str)

        lineFirstRewardFound = (
            "07:39:12.978 full_node chia.full_node.full_node: "
            + "INFO     🍀 ️Farmed unfinished_block "
            + "6b2a9249ec4aa159c24498a00305012772d33e68a01d116c1110091f440e0cf6"
        )
        lineSecondRewardFound = (
            "07:39:12.978 full_node chia.full_node.full_node: "
            + "INFO     🍀 ️Farmed unfinished_block "
            + "6b2a9249ec4aa159c24498a00305012772d33e68a01d116c1110091f440e0cf7"
        )

        self.assertTrue(action1.is_match(lineFirstRewardFound))
        self.assertTrue(action1.is_match(lineSecondRewardFound))

        chia_dog = ChiaWatchdog("", "")
        chia_dog.harvester_infos = {
            node_id: FarmerHarvesterLogfile(
                node_id,
                ip_address,
                True,
                timestamp,
            )
        }

        self.assertEqual(len(chia_dog.farmed_blocks), 0)
        action1.apply(lineFirstRewardFound, chia_dog)
        self.assertEqual(len(chia_dog.farmed_blocks), 1)
        action1.apply(lineSecondRewardFound, chia_dog)
        self.assertEqual(len(chia_dog.farmed_blocks), 2)
        action1.apply(lineSecondRewardFound, chia_dog)
        self.assertEqual(len(chia_dog.farmed_blocks), 2)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chia_tea/protobuf/generated/hardware.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='chia_tea/protobuf/generated/hardware.proto',
  package='chia_tea.protobuf.generated.hardware_pb2',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n*chia_tea/protobuf/generated/hardware.proto\x12(chia_tea.protobuf.generated.hardware_pb2\"^\n\x03\x43pu\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x63lock_speed\x18\x02 \x01(\x01\x12\r\n\x05usage\x18\x03 \x01(\x01\x12\x13\n\x0btemperature\x18\x04 \x01(\x01\x12\x10\n\x08n_vcores\x18\x05 \x01(\x05\"Q\n\x03Ram\x12\x11\n\ttotal_ram\x18\x01 \x01(\x03\x12\x10\n\x08used_ram\x18\x02 \x01(\x03\x12\x12\n\ntotal_swap\x18\x03 \x01(\x03\x12\x11\n\tused_swap\x18\x04 \x01(\x03\"\xb2\x02\n\x04\x44isk\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0btotal_space\x18\x03 \x01(\x01\x12\x12\n\nused_space\x18\x04 \x01(\x01\x12\x0e\n\x06\x64\x65vice\x18\x0c \x01(\t\x12\x12\n\nmountpoint\x18\r \x01(\t\x12\x0e\n\x06\x66stype\x18\x0e \x01(\t\x12\x15\n\rmount_options\x18\x0f \x01(\t\x12\x13\n\x0btemperature\x18\x05 \x01(\x01\x12\x15\n\rread_activity\x18\x06 \x01(\x01\x12\x16\n\x0ewrite_activity\x18\x07 \x01(\x01\x12\x12\n\nread_speed\x18\x08 \x01(\x01\x12\x13\n\x0bwrite_speed\x18\t \x01(\x01\x12\x16\n\x0eread_total_tbw\x18\n \x01(\x01\x12\x17\n\x0fwrite_total_tbw\x18\x0b \x01(\x01\x62\x06proto3'
)




_CPU = _descriptor.Descriptor(
  name='Cpu',
  full_name='chia_tea.protobuf.generated.hardware_pb2.Cpu',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='chia_tea.protobuf.generated.hardware_pb2.Cpu.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clock_speed', full_name='chia_tea.protobuf.generated.hardware_pb2.Cpu.clock_speed', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='usage', full_name='chia_tea.protobuf.generated.hardware_pb2.Cpu.usage', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='temperature', full_name='chia_tea.protobuf.generated.hardware_pb2.Cpu.temperature', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='n_vcores', full_name='chia_tea.protobuf.generated.hardware_pb2.Cpu.n_vcores', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=88,
  serialized_end=182,
)


_RAM = _descriptor.Descriptor(
  name='Ram',
  full_name='chia_tea.protobuf.generated.hardware_pb2.Ram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='total_ram', full_name='chia_tea.protobuf.generated.hardware_pb2.Ram.total_ram', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_ram', full_name='chia_tea.protobuf.generated.hardware_pb2.Ram.used_ram', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='total_swap', full_name='chia_tea.protobuf.generated.hardware_pb2.Ram.total_swap', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_swap', full_name='chia_tea.protobuf.generated.hardware_pb2.Ram.used_swap', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=184,
  serialized_end=265,
)


_DISK = _descriptor.Descriptor(
  name='Disk',
  full_name='chia_tea.protobuf.generated.hardware_pb2.Disk',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='total_space', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.total_space', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_space', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.used_space', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.device', index=4,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mountpoint', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.mountpoint', index=5,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fstype', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.fstype', index=6,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mount_options', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.mount_options', index=7,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='temperature', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.temperature', index=8,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_activity', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.read_activity', index=9,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='write_activity', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.write_activity', index=10,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_speed', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.read_speed', index=11,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='write_speed', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.write_speed', index=12,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_total_tbw', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.read_total_tbw', index=13,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='write_total_tbw', full_name='chia_tea.protobuf.generated.hardware_pb2.Disk.write_total_tbw', index=14,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=268,
  serialized_end=574,
)

DESCRIPTOR.message_types_by_name['Cpu'] = _CPU
DESCRIPTOR.message_types_by_name['Ram'] = _RAM
DESCRIPTOR.message_types_by_name['Disk'] = _DISK
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Cpu = _reflection.GeneratedProtocolMessageType('Cpu', (_message.Message,), {
  'DESCRIPTOR' : _CPU,
  '__module__' : 'chia_tea.protobuf.generated.hardware_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.hardware_pb2.Cpu)
  })
_sym_db.RegisterMessage(Cpu)

Ram = _reflection.GeneratedProtocolMessageType('Ram', (_message.Message,), {
  'DESCRIPTOR' : _RAM,
  '__module__' : 'chia_tea.protobuf.generated.hardware_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.hardware_pb2.Ram)
  })
_sym_db.RegisterMessage(Ram)

Disk = _reflection.GeneratedProtocolMessageType('Disk', (_message.Message,), {
  'DESCRIPTOR' : _DISK,
  '__module__' : 'chia_tea.protobuf.generated.hardware_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.hardware_pb2.Disk)
  })
_sym_db.RegisterMessage(Disk)


# @@protoc_insertion_point(module_scope)



import asyncio
import traceback
from typing import Callable

from ..utils.logger import get_logger
from .api.update_all import update_directly_from_chia
from .ChiaWatchdog import ChiaWatchdog
from .logfile.file_watching import watch_lines_infinitely
from .logfile.line_checks import run_line_checks
from .madmax.line_checks import run_line_checks as run_line_checks_madmax
from .regular_checks import run_watchdog_checks


async def __start_watchdog_self_checks(chia_dog: ChiaWatchdog):
    """Endless checking of the watchdog"""
    await chia_dog.ready()

    while True:
        await run_watchdog_checks(chia_dog)
        await asyncio.sleep(0.8)


def __get_on_ready_function(fn: Callable[[], None]):
    """Returns a function setting the watchog to be ready"""

    async def __on_ready():
        fn()

    return __on_ready


async def __start_updating_watchdog_service_infos(chia_dog: ChiaWatchdog):
    """Infinite loop to update the service info data of the watchdog regularly"""
    while True:
        # contact chia on same machine and update tracking data
        await update_directly_from_chia(chia_dog=chia_dog)

        # the timeout is around 4s so faster does not make sense
        await asyncio.sleep(7)


def __get_function_to_update_chia_dog_on_line(chia_dog: ChiaWatchdog):
    """Wrapper function to bring chia_dog into the context of
    the line updating function
    """

    async def _on_line_function(line: str):
        await run_line_checks(chia_dog, line)

    return _on_line_function


def __get_function_to_update_chia_dog_on_madmax_line(chia_dog: ChiaWatchdog):
    """Wrapper function to bring chia_dog into the context of
    the line updating function
    """

    async def _on_line_function(line: str):
        await run_line_checks_madmax(chia_dog, line)

    return _on_line_function


async def run_watchdog(
    chia_dog: ChiaWatchdog,
):
    """Start observing chia

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        the data of this watchdog will be updated regularly

    Notes
    -----
        Runs the following checks:
        - watches the logfile indefinitely
        - runs regular self checks
        - connects to chia services to retrieve information
    """
    while True:
        try:
            await asyncio.gather(
                # infinite watching of the chia logfile
                watch_lines_infinitely(
                    chia_dog.logfile_filepath,
                    # the watchdog is set ready if the logfile is missing
                    # this might just mean that the machine doesn't run
                    # chia.
                    on_file_missing=__get_on_ready_function(chia_dog.set_chia_logfile_is_ready),
                    on_ready=__get_on_ready_function(chia_dog.set_chia_logfile_is_ready),
                    on_line=__get_function_to_update_chia_dog_on_line(chia_dog),
                ),
                # infinite watchig of the madmax logfile
                watch_lines_infinitely(
                    chia_dog.madmax_logfile,
                    # the watchdog is set ready if the logfile is missing
                    # this might just mean that the machine doesn't run
                    # madmax.
                    on_file_missing=__get_on_ready_function(chia_dog.set_madmax_logfile_is_ready),
                    on_ready=__get_on_ready_function(chia_dog.set_madmax_logfile_is_ready),
                    on_line=__get_function_to_update_chia_dog_on_madmax_line(chia_dog),
                ),
                # regular checks such as time out
                __start_watchdog_self_checks(chia_dog),
                # regular status update from chia services
                __start_updating_watchdog_service_infos(chia_dog),
            )

        except Exception:
            err_msg = traceback.format_exc()
            get_logger(__file__).error(err_msg)
            await asyncio.sleep(5)



from datetime import datetime
from dataclasses import dataclass


class MadMaxPercentages:
    """This class contains the "rough" progress percentages for plotting

    In the future these percentages will be dynamically computed
    and estimated.
    """

    # pylint: disable=too-few-public-methods

    phase1 = (
        0.02232820817,
        0.08422978293,
        0.172304759,
        0.2580199571,
        0.3384422125,
        0.4093906676,
        0.4595129564,
    )
    phase2 = (
        0.4671522489,
        0.4938333946,
        0.5033927017,
        0.5267895266,
        0.5361274338,
        0.5582587301,
        0.5674848834,
        0.5905356127,
        0.6002520189,
        0.6219090798,
        0.6310177749,
        0.6531890657,
    )
    phase3 = (
        0.67627183,
        0.7014293004,
        0.7286114592,
        0.755564902,
        0.7815212812,
        0.8093481191,
        0.8359168967,
        0.8643466994,
        0.892014296,
        0.9206682944,
        0.9560488983,
        0.9878807454,
    )
    phase4 = 1


@dataclass
class MadMaxPlotInProgress:
    """This class represents a plot in progress created by the madmax plotter"""

    # pylint: disable=too-few-public-methods, too-many-instance-attributes

    process_id: int
    public_key: str
    pool_public_key: str
    farmer_public_key: str
    start_time: datetime
    progress: float
    plot_type: int
    state: str



from datetime import datetime
from typing import List

from ...protobuf.generated.chia_pb2 import HarvesterViewedFromFarmer
from ...protobuf.generated.computer_info_pb2 import ComputerInfo
from ...protobuf.generated.machine_info_pb2 import MachineInfo
from ..formatting import get_machine_info_name


def notify_on_harvester_reward_found(
    machine: MachineInfo,
    old_computer_info: ComputerInfo,
    new_computer_info: ComputerInfo,
) -> List[str]:
    """notify when a reward was found

    Parameters
    ----------
    machine : MachineInfo
        latest information about the machine
    old_computer_info : ComputerInfo
        computer info before last update
    new_computer_info : ComputerInfo
        computer info after last update

    Returns
    -------
    messages : List[str]
        notification messages
    """

    messages = []

    old_n_proofs = old_computer_info.harvester.n_proofs
    new_n_proofs = new_computer_info.harvester.n_proofs

    if old_n_proofs < new_n_proofs:
        messages.append(
            "{icon}   Harvester {machine_name} {status}.".format(
                icon="🔍",
                machine_name=get_machine_info_name(machine),
                status="found a proof",
            )
        )

    return messages


HARVESTER_TIMOUT = 60  # seconds
timestamp_of_last_timeout_check = 0.0


def get_msg_if_farmer_harvester_timed_out(
    machine: MachineInfo,
    last_timestamp: float,
    new_timestamp: float,
    harvester: HarvesterViewedFromFarmer,
) -> str:
    """Returns a message if a farmer recognizes that a harvester
    didn't respond for a while

    Parameters
    ----------
    machine : MachineInfo
        info about the machine itself
    last_timestamp : float
        timestamp when the last time this check was performed
    new_timestamp : float
        timestamp now
    harvester : HarvesterViewedFromFarmer
        harvester to perform timeout check on

    Returns
    -------
    msg : str
        Timeout message or an empty string if the harvester
        didn't time out.
    """

    new_harvester_timed_out = new_timestamp - harvester.time_last_msg_received >= HARVESTER_TIMOUT
    previously_notified = (
        last_timestamp - harvester.time_last_msg_received >= HARVESTER_TIMOUT
        # we assume on startup that we already notified on a timeout
        # otherwise we can a message all the time when we restart
        # the bot.
        if last_timestamp != 0.0
        else True
    )

    if new_harvester_timed_out and not previously_notified:
        return "{icon}  Harvester {machine_name} {status}.".format(
            icon="⚠️",
            machine_name=get_machine_info_name(machine),
            status=f"didn't respond for {HARVESTER_TIMOUT}s",
        )

    return ""


def notify_when_harvester_times_out(
    machine: MachineInfo,
    old_computer_info: ComputerInfo,
    new_computer_info: ComputerInfo,
) -> List[str]:
    """notify when a harvester times out

    Parameters
    ----------
    machine_id : str
        id of the machine
    old_computer_info : ComputerInfo
        computer info before last update
    new_computer_info : ComputerInfo
        computer info after last update

    Returns
    -------
    messages : List[str]
        notification messages
    """
    # pylint: disable=unused-argument

    messages = []

    # pylint: disable=global-statement
    global timestamp_of_last_timeout_check
    now = datetime.now().timestamp()

    for harvester in new_computer_info.farmer_harvesters:
        msg = get_msg_if_farmer_harvester_timed_out(
            machine,
            last_timestamp=timestamp_of_last_timeout_check,
            new_timestamp=now,
            harvester=harvester,
        )
        if msg:
            messages.append(msg)

    timestamp_of_last_timeout_check = now

    return messages


def notify_on_wallet_sync_change(
    machine: MachineInfo,
    old_computer_info: ComputerInfo,
    new_computer_info: ComputerInfo,
) -> List[str]:
    """notify when a wallet is not synced anymore

    Parameters
    ----------
    machine : MachineInfo
        latest information about the machine
    old_computer_info : ComputerInfo
        computer info before last update
    new_computer_info : ComputerInfo
        computer info after last update

    Returns
    -------
    messages : List[str]
        notification messages
    """

    messages = []

    old_wallet_synced = old_computer_info.wallet.is_synced
    new_wallet_synced = new_computer_info.wallet.is_synced

    if not old_wallet_synced and new_wallet_synced:
        messages.append(
            "{machine_name}: ✔️ The wallet 👛 has synced.".format(
                machine_name=get_machine_info_name(machine),
            )
        )
    elif old_wallet_synced and not new_wallet_synced:
        messages.append(
            "{machine_name}: ⚠ The wallet 👛 is not synced anymore".format(
                machine_name=get_machine_info_name(machine),
            )
        )

    return messages


def notify_on_wallet_connection_change(
    machine: MachineInfo,
    old_computer_info: ComputerInfo,
    new_computer_info: ComputerInfo,
) -> List[str]:
    """notify when a wallet connects or disconnects

    Parameters
    ----------
    machine : MachineInfo
        latest information about the machine
    old_computer_info : ComputerInfo
        computer info before last update
    new_computer_info : ComputerInfo
        computer info after last update

    Returns
    -------
    messages : List[str]
        notification messages
    """

    messages = []

    old_wallet_connected = old_computer_info.wallet.is_running
    new_wallet_connected = new_computer_info.wallet.is_running

    if not old_wallet_connected and new_wallet_connected:
        messages.append(
            "{machine_name}: ✔️ Connected to wallet 👛".format(
                machine_name=get_machine_info_name(machine),
            )
        )
    elif old_wallet_connected and not new_wallet_connected:
        messages.append(
            "{machine_name}: Lost connection to wallet 👛".format(
                machine_name=get_machine_info_name(machine),
            )
        )

    return messages


def get_computer_info_messages_if_any(
    machine: MachineInfo,
    old_computer_info: ComputerInfo,
    new_computer_info: ComputerInfo,
) -> List[str]:
    """Checks if we need to print any messages regarding harvesters

    Parameters
    ----------
    machine : MachineInfo
        info about the machine
    old_computer_info : ComputerInfo
        computer info before last update
    new_computer_info : ComputerInfo
        computer info after last update

    Returns
    -------
    messages : List[str]
        list of messages to print
    """
    messages = []

    for get_msg_fun in ALL_EVENT_OBSERVERS:
        messages += get_msg_fun(machine, old_computer_info, new_computer_info)

    return messages


ALL_EVENT_OBSERVERS = (
    notify_on_harvester_reward_found,
    notify_on_wallet_connection_change,
    notify_on_wallet_sync_change,
    notify_when_harvester_times_out,
)



from typing import Any, Iterable

from sortedcontainers import SortedSet

from ..protobuf.generated.computer_info_pb2 import (
    _UPDATEEVENT,
    ADD,
    DELETE,
    UPDATE,
    ComputerInfo,
    UpdateEvent,
)


def get_event_type(old_msg: Any, new_msg: Any) -> int:
    """Get the event type depending on the objects

    Parameters
    ----------
    old_msg : Any
        previous state
    new_msg : Any
        new state of the proto message

    Returns
    -------
    event_type : int
        which event type (0=ADD, 1=UPDATE, 2=DELETE)
    """
    if old_msg is None and new_msg is not None:
        return ADD
    if old_msg is not None and new_msg is not None:
        return UPDATE
    return DELETE


async def compare_computer_info(
    old_computer_info: ComputerInfo,
    new_computer_info: ComputerInfo,
) -> UpdateEvent:
    """Compares to computer infos and emits events of deltas

    Parameters
    ----------
    old_computer_info : ComputerInfo
        first computer info for comparison
    new_computer_info : ComputerInfo
        second computer info for comparison

    Yields
    ------
    udpate_event : UpdateEvent
        change events from old to new
    """

    fields = []
    if old_computer_info is not None:
        fields = old_computer_info.DESCRIPTOR.fields
    if new_computer_info is not None:
        fields = new_computer_info.DESCRIPTOR.fields

    # iterate through computer sub-messages and compare them
    for field in fields:

        # we obviously ignore meta attributes
        if field.name in ("timestamp", "machine_id"):
            continue

        # get instance members
        old_msg_or_list = getattr(old_computer_info, field.name)
        new_msg_or_list = getattr(new_computer_info, field.name)

        # bot are already messages ready to be compared
        if not isinstance(old_msg_or_list, Iterable) and not isinstance(new_msg_or_list, Iterable):
            old_msg = old_msg_or_list
            new_msg = new_msg_or_list

            if old_msg != new_msg:
                yield create_update_event(
                    old_msg=old_msg,
                    new_msg=new_msg,
                    event_type=get_event_type(old_msg_or_list, new_msg_or_list),
                )
        # both are lists of messages
        elif isinstance(old_msg_or_list, Iterable) and isinstance(new_msg_or_list, Iterable):
            old_messages = {msg.id: msg for msg in old_msg_or_list}
            new_messages = {msg.id: msg for msg in new_msg_or_list}
            all_ids = SortedSet(old_messages.keys()) | SortedSet(new_messages.keys())

            for msg_id in all_ids:
                old_msg = old_messages.get(msg_id)
                new_msg = new_messages.get(msg_id)
                if old_msg != new_msg:
                    yield create_update_event(
                        old_msg=old_msg,
                        new_msg=new_msg,
                        event_type=get_event_type(old_msg, new_msg),
                    )


def create_update_event(old_msg: Any, new_msg: Any, event_type: int) -> UpdateEvent:
    """Creates an update event from the event data

    Parameters
    ----------
    old_msg : Any
        protobuf message object at previous state
    new_msg : Any
        protobuf message object at current state
    event_type : EventType
        whether data was added, updated or removed

    Returns
    -------
    update_event : UpdateEvent
        the event ready to be sent to the server
    """
    event_data = new_msg if event_type != DELETE else old_msg

    for field in _UPDATEEVENT.fields:
        if field.message_type == event_data.DESCRIPTOR:
            kwargs = {"event_type": event_type, field.name: event_data}
            return UpdateEvent(**kwargs)

    err_msg = (
        "Could not send update event since the data"
        + " didn't match any type which can be send"
        + f" by the message protocol. Event type was '{event_type}'"
        + f" and data was of type: {type(event_data)}"
    )
    raise RuntimeError(err_msg)



from typing import List

from ..common import catch_errors_as_message, open_database_read_only
from ..formatting import get_machine_info_name
from ..notifications.run_notifiers import get_current_computer_and_machine_infos_from_db


@catch_errors_as_message
async def wallets_cmd(db_filepath: str) -> List[str]:
    """Formats all wallets into a message

    Parameters
    ----------
    db_filepath : str
        path to the sqlite database

    Returns
    -------
    messages : List[str]
        list of messages to print
    """
    messages = []

    with open_database_read_only(db_filepath) as cursor:
        machine_and_computer_info_dict = get_current_computer_and_machine_infos_from_db(cursor)
        messages = []

        for _, (machine, computer_info) in machine_and_computer_info_dict.items():
            wallet = computer_info.wallet
            if wallet.is_running:
                icon = "🟢" if wallet.is_synced else "🟠"
                not_msg = "" if wallet.is_synced else "not "
                messages.append(
                    "\n".join(
                        (
                            f"Wallet 👛 *{get_machine_info_name(machine)}*",
                            f"   {icon} {not_msg}synchronized",
                        )
                    )
                )

        # Heading in case there is anything to report
        if not messages:
            messages.append("No wallets 👛 found.")

    return messages



from datetime import datetime
from typing import List

from ...protobuf.to_sqlite.sql_cmds import (
    get_cpu_for_machine_from_db,
    get_machine_infos_from_db,
    get_ram_for_machine_from_db,
)
from ..common import catch_errors_as_message, open_database_read_only
from ..formatting import cpu_pb2_as_markdown, get_machine_info_name, ram_pb2_as_markdown


@catch_errors_as_message
async def machines_cmd(db_filepath: str) -> List[str]:
    """Formats all machines into a message

    Parameters
    ----------
    db_filepath : str
        path to the sqlite database

    Returns
    -------
    messages : List[str]
        list of messages to print
    """
    messages = []

    with open_database_read_only(db_filepath) as cursor:

        # get all machine infos from the database
        machine_infos, _ = get_machine_infos_from_db(cursor)

        # convert the infos into messages
        for machine_info in machine_infos:
            seconds_since_last_contact = (
                datetime.now() - datetime.fromtimestamp(machine_info.time_last_msg)
            ).total_seconds()
            is_connected = seconds_since_last_contact < 60
            icon = "🟢" if is_connected else "🟠"

            cpus, _ = get_cpu_for_machine_from_db(
                cursor,
                machine_info.machine_id,
            )
            cpu_msgs = [cpu_pb2_as_markdown(cpu) for cpu in cpus]

            rams, _ = get_ram_for_machine_from_db(
                cursor,
                machine_info.machine_id,
            )
            ram_msgs = [ram_pb2_as_markdown(ram) for ram in rams]

            machine_name = get_machine_info_name(machine_info)
            headline = (
                "{icon} {machine_name} responded {seconds_since_last_contact:.1f} seconds ago"
            )
            messages.append(
                "\n".join(
                    [
                        headline.format(
                            icon=icon,
                            machine_name=machine_name,
                            seconds_since_last_contact=seconds_since_last_contact,
                        ),
                    ]
                    + cpu_msgs
                    + ram_msgs
                )
            )

        # Heading in case there is anything to report
        if not messages:
            messages.append("No machines being monitored 😴")

    return messages



import sqlite3
import traceback
from contextlib import contextmanager
from typing import Generator, List


def catch_errors_as_message(function):
    """Catches errors as a list of messages

    Parameters
    ----------
    function : Coroutine[Any, Any, List[str]]
        Function to wrap. In case of an error
        the message is returned.
    """

    async def wrapper(*args, **kwargs) -> List[str]:
        try:
            return await function(*args, **kwargs)
        except Exception:
            trace = traceback.format_exc()
            return [trace]

    return wrapper


@contextmanager
def open_database_read_only(
    db_filepath: str,
) -> Generator[sqlite3.dbapi2.Cursor, None, None]:
    """Open a sqlite database read only

    Parameters
    ----------
    db_filepath : str
        filepath to database

    Returns
    -------
    connection : sqlite3.Connection
        connection object
    cursor : sqlite3.Cursor
        sqlite cursor
    """
    try:
        connection = sqlite3.connect(
            f"file:{db_filepath}?mode=ro",
            uri=True,
        )
        cursor = connection.cursor()
        yield cursor
    finally:
        connection.commit()
        connection.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chia_tea/protobuf/generated/chia.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='chia_tea/protobuf/generated/chia.proto',
  package='chia_tea.protobuf.generated.chia_pb2',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&chia_tea/protobuf/generated/chia.proto\x12$chia_tea.protobuf.generated.chia_pb2\"\xd8\x01\n\x07Process\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\nexecutable\x18\x02 \x01(\t\x12\x0f\n\x07\x63ommand\x18\x03 \x01(\t\x12\x13\n\x0b\x63reate_time\x18\x04 \x01(\x01\x12\n\n\x02id\x18\x05 \x01(\x03\x12\x11\n\tcpu_usage\x18\x06 \x01(\x02\x12\x19\n\x11used_physical_ram\x18\x07 \x01(\x02\x12\x18\n\x10used_virtual_ram\x18\x08 \x01(\x02\x12\x14\n\x0copened_files\x18\t \x01(\t\x12\x1b\n\x13network_connections\x18\n \x01(\t\"\xc4\x01\n\rHarvesterPlot\x12\n\n\x02id\x18\n \x01(\t\x12\x10\n\x08\x66ilename\x18\x03 \x01(\t\x12\x10\n\x08\x66ilesize\x18\x04 \x01(\x03\x12!\n\x19pool_contract_puzzle_hash\x18\x06 \x01(\t\x12\x17\n\x0fpool_public_key\x18\x07 \x01(\t\x12\x0c\n\x04size\x18\x08 \x01(\x03\x12\x15\n\rtime_modified\x18\t \x01(\x01\x12\x11\n\tplot_seed\x18\x01 \x01(\t\x12\x0f\n\x07\x64isk_id\x18\x02 \x01(\t\"1\n\tHarvester\x12\x12\n\nis_running\x18\x02 \x01(\x08\x12\x10\n\x08n_proofs\x18\x03 \x01(\x03\"\xbc\x01\n\x19HarvesterViewedFromFarmer\x12\n\n\x02id\x18\x01 \x01(\t\x12\x17\n\x0f\x63onnection_time\x18\x08 \x01(\x01\x12\x1e\n\x16time_last_msg_received\x18\x03 \x01(\x01\x12\x1a\n\x12time_last_msg_sent\x18\t \x01(\x01\x12\x12\n\nip_address\x18\n \x01(\t\x12\x19\n\x11missed_challenges\x18\x06 \x01(\x03\x12\x0f\n\x07n_plots\x18\x02 \x01(\x03\"6\n\x06\x46\x61rmer\x12\x12\n\nis_running\x18\x01 \x01(\x08\x12\x18\n\x10total_challenges\x18\x02 \x01(\x03\"/\n\x06Wallet\x12\x12\n\nis_running\x18\x01 \x01(\x08\x12\x11\n\tis_synced\x18\x02 \x01(\x08\"j\n\x0ePlotInProgress\x12\n\n\x02id\x18\x01 \x01(\t\x12\x17\n\x0fpool_public_key\x18\x02 \x01(\t\x12\x12\n\nstart_time\x18\x03 \x01(\x01\x12\x10\n\x08progress\x18\x05 \x01(\x02\x12\r\n\x05state\x18\x06 \x01(\tb\x06proto3'
)




_PROCESS = _descriptor.Descriptor(
  name='Process',
  full_name='chia_tea.protobuf.generated.chia_pb2.Process',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='chia_tea.protobuf.generated.chia_pb2.Process.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='executable', full_name='chia_tea.protobuf.generated.chia_pb2.Process.executable', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='command', full_name='chia_tea.protobuf.generated.chia_pb2.Process.command', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='chia_tea.protobuf.generated.chia_pb2.Process.create_time', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='chia_tea.protobuf.generated.chia_pb2.Process.id', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu_usage', full_name='chia_tea.protobuf.generated.chia_pb2.Process.cpu_usage', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_physical_ram', full_name='chia_tea.protobuf.generated.chia_pb2.Process.used_physical_ram', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='used_virtual_ram', full_name='chia_tea.protobuf.generated.chia_pb2.Process.used_virtual_ram', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='opened_files', full_name='chia_tea.protobuf.generated.chia_pb2.Process.opened_files', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='network_connections', full_name='chia_tea.protobuf.generated.chia_pb2.Process.network_connections', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=81,
  serialized_end=297,
)


_HARVESTERPLOT = _descriptor.Descriptor(
  name='HarvesterPlot',
  full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.id', index=0,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filename', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.filename', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filesize', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.filesize', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pool_contract_puzzle_hash', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.pool_contract_puzzle_hash', index=3,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pool_public_key', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.pool_public_key', index=4,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.size', index=5,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_modified', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.time_modified', index=6,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plot_seed', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.plot_seed', index=7,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disk_id', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterPlot.disk_id', index=8,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=300,
  serialized_end=496,
)


_HARVESTER = _descriptor.Descriptor(
  name='Harvester',
  full_name='chia_tea.protobuf.generated.chia_pb2.Harvester',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_running', full_name='chia_tea.protobuf.generated.chia_pb2.Harvester.is_running', index=0,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='n_proofs', full_name='chia_tea.protobuf.generated.chia_pb2.Harvester.n_proofs', index=1,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=498,
  serialized_end=547,
)


_HARVESTERVIEWEDFROMFARMER = _descriptor.Descriptor(
  name='HarvesterViewedFromFarmer',
  full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='connection_time', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.connection_time', index=1,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_last_msg_received', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.time_last_msg_received', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_last_msg_sent', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.time_last_msg_sent', index=3,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ip_address', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.ip_address', index=4,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='missed_challenges', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.missed_challenges', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='n_plots', full_name='chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer.n_plots', index=6,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=550,
  serialized_end=738,
)


_FARMER = _descriptor.Descriptor(
  name='Farmer',
  full_name='chia_tea.protobuf.generated.chia_pb2.Farmer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_running', full_name='chia_tea.protobuf.generated.chia_pb2.Farmer.is_running', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='total_challenges', full_name='chia_tea.protobuf.generated.chia_pb2.Farmer.total_challenges', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=740,
  serialized_end=794,
)


_WALLET = _descriptor.Descriptor(
  name='Wallet',
  full_name='chia_tea.protobuf.generated.chia_pb2.Wallet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_running', full_name='chia_tea.protobuf.generated.chia_pb2.Wallet.is_running', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_synced', full_name='chia_tea.protobuf.generated.chia_pb2.Wallet.is_synced', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=796,
  serialized_end=843,
)


_PLOTINPROGRESS = _descriptor.Descriptor(
  name='PlotInProgress',
  full_name='chia_tea.protobuf.generated.chia_pb2.PlotInProgress',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='chia_tea.protobuf.generated.chia_pb2.PlotInProgress.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pool_public_key', full_name='chia_tea.protobuf.generated.chia_pb2.PlotInProgress.pool_public_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='chia_tea.protobuf.generated.chia_pb2.PlotInProgress.start_time', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='progress', full_name='chia_tea.protobuf.generated.chia_pb2.PlotInProgress.progress', index=3,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='chia_tea.protobuf.generated.chia_pb2.PlotInProgress.state', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=845,
  serialized_end=951,
)

DESCRIPTOR.message_types_by_name['Process'] = _PROCESS
DESCRIPTOR.message_types_by_name['HarvesterPlot'] = _HARVESTERPLOT
DESCRIPTOR.message_types_by_name['Harvester'] = _HARVESTER
DESCRIPTOR.message_types_by_name['HarvesterViewedFromFarmer'] = _HARVESTERVIEWEDFROMFARMER
DESCRIPTOR.message_types_by_name['Farmer'] = _FARMER
DESCRIPTOR.message_types_by_name['Wallet'] = _WALLET
DESCRIPTOR.message_types_by_name['PlotInProgress'] = _PLOTINPROGRESS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Process = _reflection.GeneratedProtocolMessageType('Process', (_message.Message,), {
  'DESCRIPTOR' : _PROCESS,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.Process)
  })
_sym_db.RegisterMessage(Process)

HarvesterPlot = _reflection.GeneratedProtocolMessageType('HarvesterPlot', (_message.Message,), {
  'DESCRIPTOR' : _HARVESTERPLOT,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.HarvesterPlot)
  })
_sym_db.RegisterMessage(HarvesterPlot)

Harvester = _reflection.GeneratedProtocolMessageType('Harvester', (_message.Message,), {
  'DESCRIPTOR' : _HARVESTER,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.Harvester)
  })
_sym_db.RegisterMessage(Harvester)

HarvesterViewedFromFarmer = _reflection.GeneratedProtocolMessageType('HarvesterViewedFromFarmer', (_message.Message,), {
  'DESCRIPTOR' : _HARVESTERVIEWEDFROMFARMER,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.HarvesterViewedFromFarmer)
  })
_sym_db.RegisterMessage(HarvesterViewedFromFarmer)

Farmer = _reflection.GeneratedProtocolMessageType('Farmer', (_message.Message,), {
  'DESCRIPTOR' : _FARMER,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.Farmer)
  })
_sym_db.RegisterMessage(Farmer)

Wallet = _reflection.GeneratedProtocolMessageType('Wallet', (_message.Message,), {
  'DESCRIPTOR' : _WALLET,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.Wallet)
  })
_sym_db.RegisterMessage(Wallet)

PlotInProgress = _reflection.GeneratedProtocolMessageType('PlotInProgress', (_message.Message,), {
  'DESCRIPTOR' : _PLOTINPROGRESS,
  '__module__' : 'chia_tea.protobuf.generated.chia_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.chia_pb2.PlotInProgress)
  })
_sym_db.RegisterMessage(PlotInProgress)


# @@protoc_insertion_point(module_scope)



import asyncio
from datetime import datetime

from ...chia_watchdog.ChiaWatchdog import ChiaWatchdog
from ...protobuf.generated.computer_info_pb2 import ComputerInfo
from ...utils.logger import log_runtime_async
from .chia import (
    collect_connected_harvesters_to_farmer,
    collect_farmer_info,
    collect_harvester_info,
    collect_harvester_plots,
    collect_plots_in_progress,
    collect_process_info,
    collect_wallet_info,
)
from .hardware import collect_cpu_info, collect_disk_info, collect_ram_info


@log_runtime_async(__file__)
async def collect_computer_info(machine_id: str, chia_dog: ChiaWatchdog) -> ComputerInfo:
    """Collects all the info about the machine

    Parameters
    ----------
    machine_id : str
        id of the machine
    chia_dog : ChiaWatchdog
        chia watchdog to take data from

    Returns
    -------
    comuter_info : ComputerInfo
        info about the machine and it's chia related processes
    """

    (
        cpu_info,
        disks,
        ram_info,
        farmer_info,
        harvester_info,
        harvester_plots,
        wallet_info,
        chia_processes,
        connected_harvesters,
        plots_in_progress,
    ) = await asyncio.gather(
        collect_cpu_info(),
        collect_disk_info(),
        collect_ram_info(),
        collect_farmer_info(chia_dog),
        collect_harvester_info(chia_dog),
        collect_harvester_plots(chia_dog),
        collect_wallet_info(chia_dog),
        collect_process_info(),
        collect_connected_harvesters_to_farmer(chia_dog),
        collect_plots_in_progress(chia_dog),
    )

    computer_info = ComputerInfo(
        timestamp=datetime.now().timestamp(),
        machine_id=machine_id,
        cpu=cpu_info,
        disks=disks,
        ram=ram_info,
        farmer_harvesters=connected_harvesters,
        farmer=farmer_info,
        harvester=harvester_info,
        harvester_plots=harvester_plots,
        wallet=wallet_info,
        processes=chia_processes,
        plotting_plots=plots_in_progress,
    )

    return computer_info



from chia.rpc.wallet_rpc_client import WalletRpcClient
from chia.util.config import load_config
from chia.util.default_root import DEFAULT_ROOT_PATH
from chia.util.ints import uint16

from ..ChiaWatchdog import ChiaWatchdog
from .shared_settings import API_EXCEPTIONS


async def update_from_wallet(chia_dog: ChiaWatchdog):
    """Updates the chia dog with wallet data

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        watchdog instance to be modified
    """

    try:
        config = load_config(DEFAULT_ROOT_PATH, "config.yaml", exit_on_error=False)
        self_hostname = config["self_hostname"]
        wallet_rpc_port = config["wallet"]["rpc_port"]

        wallet_client = await WalletRpcClient.create(
            self_hostname, uint16(wallet_rpc_port), DEFAULT_ROOT_PATH, config
        )

        chia_dog.wallet_service.n_wallets = len(await wallet_client.get_connections())
        chia_dog.wallet_service.is_running = True
        chia_dog.wallet_service.is_synced = await wallet_client.get_synced()

    # pylint: disable=catching-non-exception
    except API_EXCEPTIONS:
        chia_dog.wallet_service.n_wallets = 0
        chia_dog.wallet_service.is_running = False
        chia_dog.wallet_service.is_synced = False
    finally:
        if "wallet_client" in locals():
            wallet_client.close()
            await wallet_client.await_closed()
        chia_dog.wallet_service.is_ready = True



from dataclasses import dataclass, field
from typing import Iterable, List


@dataclass
class HarvesterAPI:
    """This class holds chia information fetched through RPC
    from chia services on the same machine
    """

    # pylint: disable=too-few-public-methods

    # This helps us to check if an update happened
    is_ready: bool = False

    is_running: bool = False
    plots: List[dict] = field(default_factory=list)
    # Plot layout:
    # {
    #     'file_size': 108878195752,
    #     'filename': 'path/to/blabla.plot',
    #     'plot-seed': '0x00000000...',
    #     'plot_public_key': '0x00000000....',
    #     'pool_contract_puzzle_hash': None,
    #     'pool_public_key': '0x00000000....',
    #     'size': 32,
    #     'time_modified': 1621370658.446281
    # }
    failed_to_open_filenames: List[str] = field(default_factory=list)
    not_found_filenames: List[str] = field(default_factory=list)
    plot_directories: Iterable[str] = tuple()
    n_proofs: int = 0



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chia_tea/protobuf/generated/machine_info.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='chia_tea/protobuf/generated/machine_info.proto',
  package='chia_tea.protobuf.generated.machine_info_pb2',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n.chia_tea/protobuf/generated/machine_info.proto\x12,chia_tea.protobuf.generated.machine_info_pb2\"Z\n\x0bMachineInfo\x12\x12\n\nmachine_id\x18\x01 \x01(\x03\x12\x12\n\nip_address\x18\x02 \x01(\t\x12\x15\n\rtime_last_msg\x18\x03 \x01(\x01\x12\x0c\n\x04name\x18\x04 \x01(\tb\x06proto3'
)




_MACHINEINFO = _descriptor.Descriptor(
  name='MachineInfo',
  full_name='chia_tea.protobuf.generated.machine_info_pb2.MachineInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='machine_id', full_name='chia_tea.protobuf.generated.machine_info_pb2.MachineInfo.machine_id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ip_address', full_name='chia_tea.protobuf.generated.machine_info_pb2.MachineInfo.ip_address', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_last_msg', full_name='chia_tea.protobuf.generated.machine_info_pb2.MachineInfo.time_last_msg', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='chia_tea.protobuf.generated.machine_info_pb2.MachineInfo.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=96,
  serialized_end=186,
)

DESCRIPTOR.message_types_by_name['MachineInfo'] = _MACHINEINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MachineInfo = _reflection.GeneratedProtocolMessageType('MachineInfo', (_message.Message,), {
  'DESCRIPTOR' : _MACHINEINFO,
  '__module__' : 'chia_tea.protobuf.generated.machine_info_pb2'
  # @@protoc_insertion_point(class_scope:chia_tea.protobuf.generated.machine_info_pb2.MachineInfo)
  })
_sym_db.RegisterMessage(MachineInfo)


# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from chia_tea.protobuf.generated import computer_info_pb2 as chia__tea_dot_protobuf_dot_generated_dot_computer__info__pb2
from chia_tea.protobuf.generated import monitoring_service_pb2 as chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class MonitoringStub(object):
    """Monitoring service description with methods
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetMachineState = channel.unary_unary(
                '/chia_tea.protobuf.generated.monitoring_service_pb2.Monitoring/GetMachineState',
                request_serializer=chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2.GetStateRequest.SerializeToString,
                response_deserializer=chia__tea_dot_protobuf_dot_generated_dot_computer__info__pb2.ComputerInfo.FromString,
                )
        self.SendMonitoringUpdate = channel.stream_stream(
                '/chia_tea.protobuf.generated.monitoring_service_pb2.Monitoring/SendMonitoringUpdate',
                request_serializer=chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2.DataUpdateRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class MonitoringServicer(object):
    """Monitoring service description with methods
    """

    def GetMachineState(self, request, context):
        """This method allows to poll the current or last known state of a machine in the form of
        a ComputerInfo. This is done after establishing a connection by every client so that 
        the clients can compute the correct updates for the database to this last known state. 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendMonitoringUpdate(self, request_iterator, context):
        """SendMonitoringUpdate allows the client to send an endless stream of data update requests
        which contain UpdateEvents to be stored in the database. The server can send empty messages
        to probe the connection status if desired (not implemented at the moment).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MonitoringServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetMachineState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetMachineState,
                    request_deserializer=chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2.GetStateRequest.FromString,
                    response_serializer=chia__tea_dot_protobuf_dot_generated_dot_computer__info__pb2.ComputerInfo.SerializeToString,
            ),
            'SendMonitoringUpdate': grpc.stream_stream_rpc_method_handler(
                    servicer.SendMonitoringUpdate,
                    request_deserializer=chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2.DataUpdateRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'chia_tea.protobuf.generated.monitoring_service_pb2.Monitoring', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Monitoring(object):
    """Monitoring service description with methods
    """

    @staticmethod
    def GetMachineState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/chia_tea.protobuf.generated.monitoring_service_pb2.Monitoring/GetMachineState',
            chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2.GetStateRequest.SerializeToString,
            chia__tea_dot_protobuf_dot_generated_dot_computer__info__pb2.ComputerInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendMonitoringUpdate(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/chia_tea.protobuf.generated.monitoring_service_pb2.Monitoring/SendMonitoringUpdate',
            chia__tea_dot_protobuf_dot_generated_dot_monitoring__service__pb2.DataUpdateRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import argparse
import sys


def str2bool(value) -> bool:
    """Converts some value from the cmd line to a boolean
    Parameters
    ----------
    value: `str` or `bool`
    Returns
    -------
    bool_value: `bool`
        value as boolean
    """

    if isinstance(value, bool):
        return value

    if value.lower() in ("yes", "true", "t", "y", "1"):
        return True

    if value.lower() in ("no", "false", "f", "n", "0"):
        return False

    raise argparse.ArgumentTypeError("Boolean value expected.")


def parse_args(name: str, description: str) -> argparse.Namespace:
    """Parse the arguments from the command line

    Parameters
    ----------
    name : str
        name of the cli
    description : str
        describe shortly what the cli does

    Returns
    -------
    args : argparse.Namespace
        parsed arguments
    """

    parser = argparse.ArgumentParser(prog=name, description=description)

    parser.add_argument("--config", default="./config.yml", help="Path to the config file")

    if len(sys.argv) < 1:
        parser.print_help()
        sys.exit(0)

    return parser.parse_args(sys.argv[1:])



from typing import Any, Dict, List

import psutil

from ...chia_watchdog.ChiaWatchdog import ChiaWatchdog
from ...protobuf.generated.chia_pb2 import (
    Farmer,
    Harvester,
    HarvesterPlot,
    HarvesterViewedFromFarmer,
    PlotInProgress,
    Process,
    Wallet,
)
from ...utils.logger import log_runtime_async


@log_runtime_async(__file__)
async def collect_plots_in_progress(chia_dog: ChiaWatchdog) -> List[PlotInProgress]:
    """Converts partially finished plots from the watchdog to protobuf

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        watchdog instance to get data from

    Returns
    -------
    plots_in_progress : List[PlotInProgress]
        plots in progress converted to protobuf
    """
    plots_in_progress = []
    for plot in chia_dog.plots_in_progress:
        plot_pb2 = PlotInProgress(
            id=plot.public_key,
            pool_public_key=plot.pool_public_key,
            start_time=plot.start_time.timestamp(),
            progress=plot.progress,
            state=plot.state,
        )
        plots_in_progress.append(plot_pb2)
    return plots_in_progress


@log_runtime_async(__file__)
async def collect_connected_harvesters_to_farmer(
    chia_dog: ChiaWatchdog,
) -> List[HarvesterViewedFromFarmer]:
    """Converts harvester farmer data from watchdog to protobuf

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        watchdog instance to get data from

    Returns
    -------
    farmer_harvesters : List[HarvesterViewedFromFarmer]
        harvester data reported by farmer in proto format
    """

    harvesters_rpc = {
        connected_harvester.node_id.hex(): connected_harvester
        for connected_harvester in chia_dog.farmer_service.connections
    }
    harvesters_logfile = chia_dog.harvester_infos

    # which harvesters do exist
    all_harvester_ids = set(harvesters_rpc.keys()) | set(harvesters_logfile.keys())

    connected_harvesters = []
    for harvester_id in all_harvester_ids:

        kwargs: Dict[str, Any] = dict(
            id=harvester_id,
        )

        # get information from chia rpc
        connected_harvester = harvesters_rpc.get(harvester_id)
        if connected_harvester is not None:
            kwargs["connection_time"] = connected_harvester.creation_time
            kwargs["ip_address"] = connected_harvester.peer_host
            kwargs["n_plots"] = connected_harvester.n_plots
            # we don't use 'last_message_time' since it shows the last incoming or
            # outgoing message so we can't distinguish really.
            # kwargs["time_last_msg_sent"] = connected_harvester.last_message_time
        else:
            # there is no harvester process running anymore
            # thus we omit the data further down from the
            # logfile since we want a delete event in the db
            continue

        # get information written to logfile
        harvester_info = harvesters_logfile.get(harvester_id)
        if harvester_info:
            # don't append harvester info if it is not connected
            # anymore
            if not harvester_info.is_connected:
                continue
            kwargs["missed_challenges"] = harvester_info.n_overdue_responses
            kwargs["time_last_msg_received"] = (
                harvester_info.time_last_incoming_msg.timestamp()
                if harvester_info.time_last_incoming_msg is not None
                else 0.0
            )
            kwargs["time_last_msg_sent"] = (
                harvester_info.time_last_outgoing_msg.timestamp()
                if harvester_info.time_last_outgoing_msg is not None
                else 0.0
            )

        connected_harvesters.append(
            HarvesterViewedFromFarmer(
                **kwargs,
            )
        )

    return connected_harvesters


@log_runtime_async(__file__)
async def collect_farmer_info(chia_dog: ChiaWatchdog) -> Farmer:
    """Collects info about the farmer

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        chia watchdog to take data from

    Returns
    -------
    farmer_info : Farmer
        info about the farmer running on the system
    """
    return Farmer(
        is_running=chia_dog.farmer_service.is_running,
        # connected_harvesters=,
        # total_challenges=,
    )


@log_runtime_async(__file__)
async def collect_harvester_info(chia_dog: ChiaWatchdog) -> Harvester:
    """Collects info about the farmer

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        chia watchdog to take data from

    Returns
    -------
    info : Harvester
        info about the harvester running on the system
    plots : List[HarvesterPlot]
        list of plots on the harvester
    """

    return Harvester(
        is_running=chia_dog.harvester_service.is_running,
        n_proofs=chia_dog.harvester_service.n_proofs,
    )


@log_runtime_async(__file__)
async def collect_harvester_plots(chia_dog: ChiaWatchdog) -> List[HarvesterPlot]:
    """Collects info about the harvester plots

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        chia watchdog to take data from

    Returns
    -------
    plots : List[HarvesterPlot]
        list of plots on the harvester
    """

    plots: List[HarvesterPlot] = []
    for plot_response in chia_dog.harvester_service.plots:
        plots.append(
            HarvesterPlot(
                id=plot_response["plot_public_key"],
                plot_seed=plot_response["plot-seed"],
                filename=plot_response["filename"],
                filesize=plot_response["file_size"],
                pool_contract_puzzle_hash=plot_response["pool_contract_puzzle_hash"],
                pool_public_key=plot_response["pool_public_key"],
                size=plot_response["size"],
                time_modified=plot_response["time_modified"],
            )
        )

    return plots


@log_runtime_async(__file__)
async def collect_wallet_info(chia_dog: ChiaWatchdog) -> Wallet:
    """Collects info about the farmer

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        chia watchdog to take data from

    Returns
    -------
    wallet_info : Wallet
        info about the wallet running on the system
    """
    return Wallet(
        is_running=chia_dog.wallet_service.is_running,
        is_synced=chia_dog.wallet_service.is_synced,
    )


@log_runtime_async(__file__)
async def collect_process_info() -> List[Process]:
    """Collect data about every chia related process
    running on the machine

    Returns
    -------
    processes : List[Process]
        List of processes
    """

    process_names_to_filter_for = (
        "chia",
        "chia_harvester",
        "chia_farmer",
        "chia_wallet",
        "chia_daemon",
        "chia_full_node",
    )

    processes: List[Process] = []

    for process in psutil.process_iter():
        if not process.name() in process_names_to_filter_for:
            continue

        try:
            meminfo = process.memory_info()

            # network_connections = [
            #     "{ip}:{port}".format(
            #         ip=connection.raddr[0],
            #         port=connection.raddr[0],
            #     )
            #     for connection in process.connections()
            # ]

            processes.append(
                Process(
                    name=process.name(),
                    executable=process.exe(),
                    command="".join(process.cmdline()),
                    create_time=process.create_time(),
                    id=process.pid,
                    cpu_usage=process.cpu_percent(),
                    used_physical_ram=meminfo.rss,
                    used_virtual_ram=meminfo.vms,
                    opened_files=", ".join(file.path for file in process.open_files()),
                    # network_connections=network_connections,
                )
            )
        except (
            psutil.NoSuchProcess,
            psutil.PermissionError,
            psutil.AccessDenied,
        ):
            pass

    return processes



from typing import List

from ..common import catch_errors_as_message, open_database_read_only
from ..formatting import get_machine_info_name, plot_in_progress_pb2_as_markdown
from ..notifications.run_notifiers import get_current_computer_and_machine_infos_from_db


@catch_errors_as_message
async def plotters_cmd(db_filepath: str) -> List[str]:
    """Formats all harvesters into a message

    Parameters
    ----------
    db_filepath : str
        path to the sqlite database

    Returns
    -------
    messages : List[str]
        list of messages to print
    """
    messages: List[str] = []

    with open_database_read_only(db_filepath) as cursor:

        machine_and_computer_info_dict = get_current_computer_and_machine_infos_from_db(cursor)

        for _, (machine, computer_info) in machine_and_computer_info_dict.items():
            if computer_info.plotting_plots:
                messages.append(
                    "\n".join(
                        (f"  🛠️ Plotter {get_machine_info_name(machine)}",)
                        + tuple(
                            plot_in_progress_pb2_as_markdown(plot)
                            for plot in computer_info.plotting_plots
                        )
                    )
                )

        if not messages:
            messages.append("No Plotters 🛠️ at work.")

    return messages



import asyncio

import aiohttp

API_EXCEPTIONS = (
    # No config or error response
    ValueError,
    # No chia on machine
    RuntimeError,
    # Not running
    aiohttp.ClientConnectorError,
    asyncio.TimeoutError,
)



import unittest

from chia.server.outbound_message import NodeType

from ...chia_watchdog.api.FarmerHarvesterAPI import FarmerHarvesterAPI
from ...chia_watchdog.ChiaWatchdog import ChiaWatchdog
from ...chia_watchdog.logfile.FarmerHarvesterLogfile import FarmerHarvesterLogfile
from ..generated.chia_pb2 import HarvesterViewedFromFarmer
from .chia import collect_connected_harvesters_to_farmer


class TestChiaDataCollection(unittest.TestCase):
    async def test_farmer_harvester_collection(self):

        node_id = b"1n\x0f\xc4J\xb5q8\xc4\x98\x0b\xe7\\\xac\xd1\x82"
        node_id_str = node_id.hex()
        plots = ["only", "length", "is", "used"]
        creation_time = 1.0
        time_last_msg_received = 2.5
        time_last_msg_sent = 1.5
        ip_address = "127.0.0.2"
        missed_challenges = 3

        harvster_api = FarmerHarvesterAPI(
            node_id=node_id,
            bytes_read=732920,
            bytes_written=736979,
            creation_time=creation_time,
            last_message_time=time_last_msg_received,
            local_port=8447,
            peer_host=ip_address,
            peer_port=51844,
            peer_server_port=8448,
            type=NodeType.HARVESTER.value,
            n_plots=len(plots),
        )
        harvester_logfile = FarmerHarvesterLogfile(
            harvester_id=node_id_str,
            ip_address=ip_address,
            last_update=max(time_last_msg_received, time_last_msg_sent),
            time_last_incoming_msg=time_last_msg_received,
            time_last_outgoing_msg=time_last_msg_sent,
            time_of_end_of_last_sgn_point=1,
            timed_out=False,
        )

        dog = ChiaWatchdog("", "")
        dog.farmer_service.connections = [harvster_api]
        dog.harvester_infos = {node_id_str: harvester_logfile}

        # do the thing
        harvester_list = await collect_connected_harvesters_to_farmer(dog)

        self.assertTrue(len(harvester_list), 1)
        harvester: HarvesterViewedFromFarmer = harvester_list[0]
        self.assertEqual(harvester.id, node_id)
        self.assertEqual(harvester.connection_time, node_id)
        self.assertEqual(harvester.time_last_msg_received, time_last_msg_received)
        self.assertEqual(harvester.time_last_msg_sent, time_last_msg_sent)
        self.assertEqual(harvester.ip_address, ip_address)
        self.assertEqual(harvester.missed_challenges, missed_challenges)
        self.assertEqual(harvester.n_plots, len(plots))



import os

import yaml
from google.protobuf.json_format import MessageToDict, ParseDict

from ..protobuf.generated.config_pb2 import (
    INFO,
    ChiaConfig,
    ChiaTeaConfig,
    CopyConfig,
    DevelopmentConfig,
    DiscordConfig,
    LoggingConfig,
    MachineConfig,
    MonitoringConfig,
)


def create_default_config(filepath: str) -> ChiaTeaConfig:
    """Creates a default config on the system

    Parameters
    ----------
    filepath : str
        path where to create the default config

    Returns
    -------
    config : ChiaTeaConfig
        created default config object

    Notes
    -----
        The config is being stored in the '.chia_tea'
        directory in the home folder.
    """
    config_filepath = os.path.join(
        filepath,
        "config.yml",
    )
    if not os.path.exists(config_filepath):
        config = get_default_config()
        save_config(config_filepath, get_default_config())
    else:
        config = get_config()

    return config


def get_default_config() -> ChiaTeaConfig:
    """Get the default config for chia tea

    Returns
    -------
    config : ChiaTeaConfig
        default config
    """

    return ChiaTeaConfig(
        version=0,
        # configuration for machine-specific settings
        machine=MachineConfig(
            name="machine_name_to_display",
        ),
        logging=LoggingConfig(
            # Allowed is: TRACE, DEBUG, INFO, WARNING, ERROR
            loglevel=INFO,
            log_to_console=True,
            log_to_file=True,
            max_logfiles=10,
            max_logfile_size_mb=15,
        ),
        # This controls from where data is copied
        # with the CLI copy tool. The copy process
        # selects disks smartly. Also it can deal
        # with disconnects and other issues.
        copy=CopyConfig(
            source_folders=[
                "/some/plotting/folder",
                "/another/plotting/folder",
            ],
            target_folders=[
                "/some/harvester/folder",
                "/another/harvester/folder",
            ],
        ),
        # General chia-related settings
        chia=ChiaConfig(
            # Filepath at which the chia logfile resides.
            logfile_filepath="~/.chia/mainnet/log/debug.log",
            # Filepath to the logfile of the madmax plotter.
            # Leave empty if unused
            madmax_logfile="",
        ),
        # Settings to connect to the discord bot and channel
        discord=DiscordConfig(
            token="YOUR_DISCORD_TOKEN",
            channel_id=123456789,
        ),
        # Specify monitoring settings
        monitoring=MonitoringConfig(
            # The communication is secured by certificates.
            # These need to be generated beforehand (see Taskfile).
            # One done you can register them here. The server needs
            # both while the client just needs the cert-file.
            auth=MonitoringConfig.AuthConfig(
                cert_filepath="./server.crt",
                key_filepath="./server.key",
            ),
            # Settings for the monitoring server
            server=MonitoringConfig.ServerConfig(
                port=43200,
                # filepath where the monitoring database will
                # be stored on disk.
                db_filepath="./monitoring.db",
            ),
            # Settings for the client collecting data
            client=MonitoringConfig.ClientConfig(
                # Address of the server to send data to.
                # On the very same machine this is 127.0.0.1.
                address="127.0.0.1",
                port=43200,
                # Depending on the system the data collection
                # can be quite fast and may use too much cpu.
                # This option throttles the general data
                # collection. If the data collection takes
                # longer than specified here the waiting time
                # is ignored.
                collect_data_every=1.5,
                # To not spam the database with too much data
                # we can set a limit here to send updates no
                # faster than specified.
                # Note that if nothing changed such as disk space
                # then no update is sent at all to keep stay
                # efficient.
                send_update_every=MonitoringConfig.ClientConfig.SendUpdateEvery(
                    # For hardware it is recommended to choose
                    # 1-5 minutes
                    cpu=60,
                    ram=60,
                    disk=60,
                    process=60,
                    # Chia data is not limited by default since
                    # we want to know asap if something is up.
                    # Note though that 'collect_data_every'
                    # applies and automatically rate limits
                    # chia data.
                ),
            ),
        ),
        # Enables development mode. This currently disables
        # encryption and also the discord bot does not send
        # send messages but they are printed to console.
        development=DevelopmentConfig(
            testing=False,
        ),
    )


def read_config(filepath: str) -> ChiaTeaConfig:
    """Reads the config file

    Parameters
    ----------
    filepath : str
        path to the config file

    Returns
    -------
    config : ChiaTeaConfig
        read config file

    Raises
    ------
    yaml.YAMLError
        In case of any issues with yaml parsing
    ValueError
        If config values are invalid
    """
    with open(filepath, "r", encoding="utf8") as stream:
        config_dict = yaml.safe_load(stream)

    config = ParseDict(
        js_dict=config_dict,
        message=ChiaTeaConfig(),
    )

    # pylint: disable=global-statement
    global __GLOBAL_CONFIG
    global __IS_LOADED
    __GLOBAL_CONFIG = config
    __IS_LOADED = True

    return config


def config_is_loaded() -> bool:
    """Checks whether a config was loaded

    Returns
    -------
    is_loaded : bool
        whether the config is loaded
    """
    return __IS_LOADED


def save_config(filepath: str, config: ChiaTeaConfig):
    """Save the config to a file

    Parameters
    ----------
    filepath : str
        path to save the config
    config : ChiaTeaConfig
        config to save
    """

    with open(filepath, "w", encoding="utf8") as file_handle:
        file_handle.write(
            yaml.dump(
                MessageToDict(
                    message=config,
                    including_default_value_fields=True,
                    use_integers_for_enums=False,
                    preserving_proto_field_name=True,
                )
            )
        )


# This is the global application config
__GLOBAL_CONFIG = get_default_config()
__IS_LOADED = False


def get_config() -> ChiaTeaConfig:
    """Get the global application config

    Returns
    -------
    config : ChiaTeaConfig
        global config
    """
    return __GLOBAL_CONFIG



import os
import tempfile
import unittest

from google.protobuf.json_format import ParseDict

from ...monitoring.MonitoringDatabase import MonitoringDatabase
from ...protobuf.generated.computer_info_pb2 import ADD, UpdateEvent
from ...protobuf.generated.monitoring_service_pb2 import DataUpdateRequest
from ...utils.testing import async_test
from .plotters import plotters_cmd


class TestPlotterCmd(unittest.TestCase):
    @async_test
    async def test_no_plotter_case(self) -> None:

        with tempfile.TemporaryDirectory() as tmpdir:
            db_filepath = os.path.join(tmpdir, "temp.db")

            with MonitoringDatabase(db_filepath):
                messages = await plotters_cmd(db_filepath)

                self.assertEqual(len(messages), 1)
                self.assertTrue(messages[0].startswith("No Plotters"))

    @async_test
    async def test_display_monitored_plotter(self) -> None:

        with tempfile.TemporaryDirectory() as tmpdir:
            db_filepath = os.path.join(tmpdir, "tmp.db")

            with MonitoringDatabase(db_filepath) as db:
                update_events = [
                    ParseDict(
                        js_dict=dict(
                            event_type=ADD,
                            plotting_plot=dict(
                                id="id",
                                pool_public_key="pool_public_key",
                                start_time=999.0,
                                progress=0.1,
                                state="Some Phase",
                            ),
                        ),
                        message=UpdateEvent(),
                    ),
                ]
                request = DataUpdateRequest(
                    machine_id=1,
                    machine_name="machine A",
                    timestamp=1000,
                    events=update_events,
                )
                db.store_data_update_request(request)

                messages = await plotters_cmd(db_filepath)
                # no failure
                self.assertEqual(len(messages), 1)
                msg = messages[0]
                self.assertFalse(msg.startswith("Traceback"))
                # display online harvester
                self.assertIn("Plot id", msg)
                self.assertIn("Since:", msg)
                self.assertIn("State: Some Phase", msg)
                self.assertIn("Progress: 10.0%", msg)

    @async_test
    async def test_fix_62_print_no_plotters_if_machines_are_present(self) -> None:

        with tempfile.TemporaryDirectory() as tmpdir:
            db_filepath = os.path.join(tmpdir, "temp.db")

            with MonitoringDatabase(db_filepath) as db:
                update_events = [
                    ParseDict(
                        js_dict=dict(
                            event_type=ADD,
                            # just add something so we have a machine
                            wallet=dict(
                                is_running=True,
                                is_synced=True,
                            ),
                        ),
                        message=UpdateEvent(),
                    ),
                ]
                request = DataUpdateRequest(
                    machine_id=1,
                    machine_name="machine A",
                    timestamp=1000,
                    events=update_events,
                )
                db.store_data_update_request(request)

            with MonitoringDatabase(db_filepath):
                messages = await plotters_cmd(db_filepath)

                self.assertEqual(len(messages), 1)
                self.assertTrue(messages[0].startswith("No Plotters"))



import unittest
from datetime import datetime
from unittest import mock

import psutil

from .ChiaWatchdog import ChiaWatchdog
from .madmax.MadMaxPlotInProgress import MadMaxPlotInProgress
from .regular_checks import remove_plotting_plots_if_madmax_does_not_run


class TestRegularChecks(unittest.TestCase):
    def test_plotting_plot_is_not_removed_if_plotting_job_is_running(self):

        watchdog = ChiaWatchdog("", "")
        watchdog.plots_in_progress = [
            MadMaxPlotInProgress(
                process_id=1,
                farmer_public_key="",
                pool_public_key="",
                public_key="",
                start_time=datetime.now(),
                progress=0,
                plot_type=32,
                state="",
            )
        ]
        with mock.patch("psutil.Process", mock.MagicMock()) as MockedProcess:
            MockedProcess = MockedProcess.return_value
            MockedProcess.is_running.return_value = True
            MockedProcess.exe.return_value = "chia_plot"
            remove_plotting_plots_if_madmax_does_not_run(watchdog)
        self.assertEqual(len(watchdog.plots_in_progress), 1)

    def test_plotting_plot_is_removed_if_plotting_job_is_gone(self):

        watchdog = ChiaWatchdog("", "")
        plot = MadMaxPlotInProgress(
            process_id=1,
            farmer_public_key="",
            pool_public_key="",
            public_key="",
            start_time=datetime.now(),
            progress=0,
            plot_type=32,
            state="",
        )
        watchdog.plots_in_progress = [plot]
        with mock.patch("psutil.Process", mock.MagicMock()) as MockedProcess:
            MockedProcess.side_effect = psutil.NoSuchProcess(100, "", "")
            remove_plotting_plots_if_madmax_does_not_run(watchdog)
        self.assertEqual(len(watchdog.plots_in_progress), 0)

        watchdog.plots_in_progress = [plot]
        with mock.patch("psutil.Process", mock.MagicMock()) as MockedProcess:
            MockedProcess = MockedProcess.return_value
            MockedProcess.is_running.return_value = True
            MockedProcess.exe.return_value = "some_other_program"
            remove_plotting_plots_if_madmax_does_not_run(watchdog)
        self.assertEqual(len(watchdog.plots_in_progress), 0)

        watchdog.plots_in_progress = [plot]
        with mock.patch("psutil.Process", mock.MagicMock()) as MockedProcess:
            MockedProcess.is_running.return_value = False
            MockedProcess.exe.return_value = "pewpew/chia_plot"
            remove_plotting_plots_if_madmax_does_not_run(watchdog)
        self.assertEqual(len(watchdog.plots_in_progress), 0)



import os
import tempfile
import unittest

from ...utils.testing import async_test
from .file_watching import watch_logfile_generator


class TestChiaWatchdog(unittest.TestCase):
    @async_test
    async def test_watch_logfile_generator_startup(self):

        some_list = []

        async def set_ready():
            some_list.append("something")

        lines = [
            "A",
            "B",
        ]

        with tempfile.TemporaryDirectory() as dir_path:
            filepath = os.path.join(dir_path, "test.log")
            with open(filepath, "w", encoding="utf-8") as fp:
                for line in lines:
                    fp.write(line + "\n")

            line_generator = watch_logfile_generator(filepath, on_ready=set_ready)

            for result_line in lines:
                line = await line_generator.asend(None)
                line = line.replace("\n", "")
                self.assertEqual(line, result_line)

            self.assertTrue(len(some_list) != 0)

    @async_test
    async def test_watch_logfile_generator_work_during_rollovers(self):

        some_list = []

        async def set_ready():
            some_list.append("something")

        lines = ["A", "B", "C"]

        with tempfile.TemporaryDirectory() as dir_path:
            filepath = os.path.join(dir_path, "test.log")
            filepath_rollover = os.path.join(dir_path, "test_1.log")
            with open(filepath, "w", encoding="utf-8") as fp:
                for line in lines[:-1]:
                    fp.write(line + "\n")

            line_generator = watch_logfile_generator(filepath, on_ready=set_ready)

            for result_line in lines[:-1]:
                line = await line_generator.asend(None)
                line = line.replace("\n", "")
                self.assertEqual(line, result_line)

            self.assertTrue(len(some_list) != 0)

            # do rollover
            os.rename(filepath, filepath_rollover)
            with open(filepath, "w", encoding="utf-8") as fp:
                fp.write(lines[-1] + "\n")

            line = await line_generator.asend(None)
            line = line.replace("\n", "")
            self.assertEqual(line, lines[-1])



from typing import Callable, List

from ...protobuf.generated.chia_pb2 import HarvesterPlot, HarvesterViewedFromFarmer
from ...protobuf.generated.computer_info_pb2 import ADD, DELETE, UpdateEvent
from ...protobuf.generated.hardware_pb2 import Disk, Ram
from ...protobuf.to_sqlite.custom import get_update_even_data


def _get_harvester_connection_msg(
    farmer_id: str,
    harvester_id: str,
    ip_address: str,
    is_connected: bool,
) -> str:
    """Get the connection msg for a harvester

    Parameters
    ----------
    farmer_id : str
        id of the farmer
    harvester_id : str
        id of the harvester
    ip_address : str
        harvester ip address
    is_connected: bool
        connection status

    Returns
    -------
    msg : str
        harvester connection message
    """

    # send message
    icon = "🟢" if is_connected else "🟠"
    connection_status = "connected" if is_connected else "disconnected"

    CONNECTION_MSG: str = "{icon} Farmer {farmer_id} {status} to Harvester {harvester_id} ({ip})."
    msg = CONNECTION_MSG.format(
        icon=icon,
        harvester_id=harvester_id[:10],
        ip=ip_address,
        status=connection_status,
        farmer_id=farmer_id,
    )

    return msg


def notify_when_farmer_connects_or_disconnects_to_harvester(
    machine_id: str,
    update_events: List[UpdateEvent],
) -> List[str]:
    """notify when a farmer connected to a farmer

    Parameters
    ----------
    machine_id : str
        id of the machine
    update_events : List[UpdateEvent]
        list of change events which happened

    Returns
    -------
    messages : List[str]
        notification messages
    """
    messages = []

    for event in update_events:
        _, pb_msg = get_update_even_data(event)
        if isinstance(pb_msg, HarvesterViewedFromFarmer):
            if event.event_type in (ADD, DELETE):
                is_connected = event.event_type == ADD
                messages.append(
                    _get_harvester_connection_msg(
                        farmer_id=str(machine_id)[:10],
                        harvester_id=pb_msg.id,
                        ip_address=pb_msg.ip_address,
                        is_connected=is_connected,
                    )
                )

    return messages


def notify_on_full_ram(
    machine_id: str,
    update_events: List[UpdateEvent],
) -> List[str]:
    """notify if the ram of a machine is full

    Parameters
    ----------
    machine_id : str
        id of the machine
    update_events : List[UpdateEvent]
        list of change events which happened

    Returns
    -------
    messages : List[str]
        notification messages
    """

    messages = []

    # check for messages
    for event in update_events:
        _, pb_msg = get_update_even_data(event)
        if isinstance(pb_msg, Ram):
            ram_usage = pb_msg.used_ram / pb_msg.total_ram
            if ram_usage > 0.95:
                messages.append(f"⚠️ Machine {machine_id} uses {ram_usage * 100:.1f}% of RAM")

    return messages


def notify_if_a_disk_is_lost(machine_id: str, update_events: List[UpdateEvent]) -> List[str]:
    """notify if a machine looses a disk

    Parameters
    ----------
    machine_id : str
        id of the machine
    update_events : List[UpdateEvent]
        list of change events which happened

    Returns
    -------
    messages : List[str]
        notification messages
    """

    messages = []
    for event in update_events:
        _, pb_msg = get_update_even_data(event)
        if isinstance(pb_msg, Disk):
            if event.event_type == DELETE:
                messages.append(f"⚠️   Machine {machine_id} lost disk {pb_msg.mountpoint}")

    return messages


def notify_if_plots_are_lost(machine_id: str, update_events: List[UpdateEvent]) -> List[str]:
    """notify if a machine looses a disk

    Parameters
    ----------
    machine_id : str
        id of the machine
    update_events : List[UpdateEvent]
        list of change events which happened

    Returns
    -------
    messages : List[str]
        notification messages
    """

    n_deleted_plots = 0
    for event in update_events:
        _, pb_msg = get_update_even_data(event)
        if isinstance(pb_msg, HarvesterPlot):
            if event.event_type == DELETE:
                n_deleted_plots += 1

    messages = []
    if n_deleted_plots:
        messages.append(f"⚠️ Machine {machine_id} lost {n_deleted_plots} plots 🌽")

    return messages


def get_update_event_messages_if_any(
    machine_id: str,
    update_events: List[UpdateEvent],
) -> List[str]:
    """Checks if we need to print any messages regarding harvesters

    Parameters
    ----------
    update_events : List[UpdateEvent]
        update events which happened

    Returns
    -------
    messages : List[str]
        list of messages to print
    """
    messages = []

    for get_msg_fun in ALL_EVENT_OBSERVERS:
        messages += get_msg_fun(machine_id, update_events)

    return messages


ALL_EVENT_OBSERVERS: Callable[[str, List[UpdateEvent]], List[str]] = (
    notify_when_farmer_connects_or_disconnects_to_harvester,
    notify_on_full_ram,
    notify_if_plots_are_lost,
    notify_if_a_disk_is_lost,
)
